# -*- coding: utf-8 -*-

from __future__ import division

import os
from collections import Counter

import numpy as np
from scipy import sparse

from pygsp2 import utils
from .fourier import FourierMixIn
from .difference import DifferenceMixIn
from ._io import IOMixIn
from ._layout import LayoutMixIn


class Graph(FourierMixIn, DifferenceMixIn, IOMixIn, LayoutMixIn):
    r"""Base graph class.

    * Instantiate it to construct a graph from a (weighted) adjacency matrix.
    * Provide a common interface (and implementation) for graph objects.
    * Initialize attributes for derived classes.

    Parameters
    ----------
    adjacency : sparse matrix or array_like
        The (weighted) adjacency matrix of size n_vertices by n_vertices that
        encodes the graph.
        The data is copied except if it is a sparse matrix in CSR format.
    lap_type : {'combinatorial', 'normalized'}
        The kind of Laplacian to be computed by :meth:`compute_laplacian`.
    coords : array_like
        A matrix of size n_vertices by d that represents the coordinates of the
        vertices in a d-dimensional embedding space.
    plotting : dict
        Plotting parameters.

    Attributes
    ----------
    n_vertices or N : int
        The number of vertices (nodes) in the graph.
    n_edges or Ne : int
        The number of edges (links) in the graph.
    W : :class:`scipy.sparse.csr_matrix`
        The adjacency matrix that contains the weights of the edges.
        It is represented as an n_vertices by n_vertices matrix, where
        :math:`W_{i,j}` is the weight of the edge :math:`(v_i, v_j)` from
        vertex :math:`v_i` to vertex :math:`v_j`. :math:`W_{i,j} = 0` means
        that there is no direct connection.
    L : :class:`scipy.sparse.csr_matrix`
        The graph Laplacian, an N-by-N matrix computed from W.
    lap_type : 'normalized', 'combinatorial'
        The kind of Laplacian that was computed by :func:`compute_laplacian`.
    signals : dict (string -> :class:`numpy.ndarray`)
        Signals attached to the graph.
    coords : :class:`numpy.ndarray`
        Vertices coordinates in 2D or 3D space. Used for plotting only.
    plotting : dict
        Plotting parameters.

    Examples
    --------

    Define a simple graph.

    >>> graph = graphs.Graph([
    ...     [0., 2., 0.],
    ...     [2., 0., 5.],
    ...     [0., 5., 0.],
    ... ])
    >>> graph
    Graph(n_vertices=3, n_edges=2)
    >>> graph.n_vertices, graph.n_edges
    (3, 2)
    >>> graph.W.toarray()
    array([[0., 2., 0.],
           [2., 0., 5.],
           [0., 5., 0.]])
    >>> graph.d
    array([1, 2, 1], dtype=int32)
    >>> graph.dw
    array([2., 7., 5.])
    >>> graph.L.toarray()
    array([[ 2., -2.,  0.],
           [-2.,  7., -5.],
           [ 0., -5.,  5.]])

    Add some coordinates to plot it.

    >>> import matplotlib.pyplot as plt
    >>> graph.set_coordinates([
    ...     [0, 0],
    ...     [0, 1],
    ...     [1, 0],
    ... ])
    >>> fig, ax = graph.plot()

    """

    def __init__(self, adjacency, lap_type='combinatorial', coords=None,
                 plotting={}):

        self.logger = utils.build_logger(__name__)

        if not sparse.isspmatrix(adjacency):
            adjacency = np.asanyarray(adjacency)

        if (adjacency.ndim != 2) or (adjacency.shape[0] != adjacency.shape[1]):
            raise ValueError('Adjacency: must be a square matrix.')

        # CSR sparse matrices are the most efficient for matrix multiplication.
        # They are the sole sparse matrix type to support eliminate_zeros().
        self._adjacency = sparse.csr_matrix(adjacency, copy=False)

        if np.isnan(self._adjacency.sum()):
            raise ValueError('Adjacency: there is a Not a Number (NaN).')
        if np.isinf(self._adjacency.sum()):
            raise ValueError('Adjacency: there is an infinite value.')
        if self.has_loops():
            self.logger.warning('Adjacency: there are self-loops '
                                '(non-zeros on the diagonal). '
                                'The Laplacian will not see them.')
        if (self._adjacency < 0).nnz != 0:
            self.logger.warning('Adjacency: there are negative edge weights.')

        self.n_vertices = self._adjacency.shape[0]

        # Don't keep edges of 0 weight. Otherwise n_edges will not correspond
        # to the real number of edges. Problematic when plotting.
        self._adjacency.eliminate_zeros()

        self._directed = None
        self._connected = None

        # Don't count edges two times if undirected.
        # Be consistent with the size of the differential operator.
        if self.is_directed():
            self.n_edges = self._adjacency.nnz
        else:
            diagonal = np.count_nonzero(self._adjacency.diagonal())
            off_diagonal = self._adjacency.nnz - diagonal
            self.n_edges = off_diagonal // 2 + diagonal

        if coords is not None:
            # TODO: self.coords should be None if unset.
            self.coords = np.asanyarray(coords)

        self.plotting = {
                'vertex_size': 100,
                'vertex_color': (0.12, 0.47, 0.71, 0.5),
                'edge_color': (0.5, 0.5, 0.5, 0.5),
                'edge_width': 2,
                'edge_style': '-',
                'highlight_color': 'C1',
                'normalize_intercept': .25,
        }
        self.plotting.update(plotting)
        self.signals = dict()

        # Attributes that are lazily computed.
        self._A = None
        self._d = None
        self._dw = None
        self._lmax = None
        self._lmax_method = None
        self._U = None
        self._e = None
        self._coherence = None
        self._D = None
        # self._L = None

        # TODO: what about Laplacian? Lazy as Fourier, or disallow change?
        self.lap_type = lap_type
        self.compute_laplacian(lap_type)

        # TODO: kept for backward compatibility.
        self.Ne = self.n_edges
        self.N = self.n_vertices

    def _get_extra_repr(self):
        return dict()

    def __repr__(self, limit=None):
        s = ''
        for attr in ['n_vertices', 'n_edges']:
            s += '{}={}, '.format(attr, getattr(self, attr))
        for i, (key, value) in enumerate(self._get_extra_repr().items()):
            if (limit is not None) and (i == limit - 2):
                s += '..., '
                break
            s += '{}={}, '.format(key, value)
        return '{}({})'.format(self.__class__.__name__, s[:-2])

    def set_signal(self, signal, name):
        r"""Attach a signal to the graph.

        Attached signals can be accessed (and modified or deleted) through the
        :attr:`signals` dictionary.

        Parameters
        ----------
        signal : array_like
            A sequence that assigns a value to each vertex.
            The value of the signal at vertex `i` is ``signal[i]``.
        name : String
            Name of the signal used as a key in the :attr:`signals` dictionary.

        Examples
        --------
        >>> graph = graphs.Sensor(10)
        >>> signal = np.arange(graph.n_vertices)
        >>> graph.set_signal(signal, 'mysignal')
        >>> graph.signals
        {'mysignal': array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])}

        """
        signal = self._check_signal(signal)
        self.signals[name] = signal

    def subgraph(self, vertices):
        r"""Create a subgraph from a list of vertices.

        Parameters
        ----------
        vertices : list
            Vertices to keep.
            Either a list of indices or an indicator function.

        Returns
        -------
        subgraph : :class:`Graph`
            Subgraph.

        Examples
        --------
        >>> graph = graphs.Graph([
        ...     [0., 3., 0., 0.],
        ...     [3., 0., 4., 0.],
        ...     [0., 4., 0., 2.],
        ...     [0., 0., 2., 0.],
        ... ])
        >>> graph = graph.subgraph([0, 2, 1])
        >>> graph.W.toarray()
        array([[0., 0., 3.],
               [0., 0., 4.],
               [3., 4., 0.]])

        """
        adjacency = self.W[vertices, :][:, vertices]
        try:
            coords = self.coords[vertices]
        except AttributeError:
            coords = None
        graph = Graph(adjacency, self.lap_type, coords, self.plotting)
        for name, signal in self.signals.items():
            graph.set_signal(signal[vertices], name)
        return graph

    def is_weighted(self):
        r"""Check if the graph is weighted.

        A graph is unweighted (binary) if and only if all the entries in the
        adjacency matrix are either zero or one.

        Returns
        -------
        weighted : bool
            True if the graph is weighted, False otherwise.

        Examples
        --------

        Unweighted (binary) graph:

        >>> graph = graphs.Graph([
        ...     [0, 1, 0],
        ...     [1, 0, 1],
        ...     [0, 1, 0],
        ... ])
        >>> graph.is_weighted()
        False

        Weighted graph:

        >>> graph = graphs.Graph([
        ...     [0, 2, 0],
        ...     [2, 0, 1],
        ...     [0, 1, 0],
        ... ])
        >>> graph.is_weighted()
        True

        """
        return not np.all(self.W.data == 1)

    def is_connected(self):
        r"""Check if the graph is connected (cached).

        A graph is connected if and only if there exists a (directed) path
        between any two vertices.

        Returns
        -------
        connected : bool
            True if the graph is connected, False otherwise.

        Notes
        -----

        For undirected graphs, starting at a vertex and trying to visit all the
        others is enough.
        For directed graphs, one needs to check that a vertex can both be
        visited by all the others and visit all the others.

        Examples
        --------

        Connected graph:

        >>> graph = graphs.Graph([
        ...     [0, 3, 0, 0],
        ...     [3, 0, 4, 0],
        ...     [0, 4, 0, 2],
        ...     [0, 0, 2, 0],
        ... ])
        >>> graph.is_connected()
        True

        Disconnected graph:

        >>> graph = graphs.Graph([
        ...     [0, 3, 0, 0],
        ...     [3, 0, 4, 0],
        ...     [0, 0, 0, 2],
        ...     [0, 0, 2, 0],
        ... ])
        >>> graph.is_connected()
        False


        """
        if self._connected is not None:
            return self._connected

        adjacencies = [self.W]
        if self.is_directed():
            adjacencies.append(self.W.T)

        for adjacency in adjacencies:
            visited = np.zeros(self.n_vertices, dtype=bool)
            stack = set([0])

            while stack:
                vertex = stack.pop()

                if visited[vertex]:
                    continue
                visited[vertex] = True

                neighbors = adjacency[vertex].nonzero()[1]
                stack.update(neighbors)

            if not np.all(visited):
                self._connected = False
                return self._connected

        self._connected = True
        return self._connected

    def is_directed(self):
        r"""Check if the graph has directed edges (cached).

        In this framework, we consider that a graph is directed if and
        only if its weight matrix is not symmetric.

        Returns
        -------
        directed : bool
            True if the graph is directed, False otherwise.

        Examples
        --------

        Directed graph:

        >>> graph = graphs.Graph([
        ...     [0, 3, 0],
        ...     [3, 0, 4],
        ...     [0, 0, 0],
        ... ])
        >>> graph.is_directed()
        True

        Undirected graph:

        >>> graph = graphs.Graph([
        ...     [0, 3, 0],
        ...     [3, 0, 4],
        ...     [0, 4, 0],
        ... ])
        >>> graph.is_directed()
        False

        """
        if self._directed is None:
            self._directed = (self.W != self.W.T).nnz != 0
        return self._directed

    def has_loops(self):
        r"""Check if any vertex is connected to itself.

        A graph has self-loops if and only if the diagonal entries of its
        adjacency matrix are not all zero.

        Returns
        -------
        loops : bool
            True if the graph has self-loops, False otherwise.

        Examples
        --------

        Without self-loops:

        >>> graph = graphs.Graph([
        ...     [0, 3, 0],
        ...     [3, 0, 4],
        ...     [0, 0, 0],
        ... ])
        >>> graph.has_loops()
        False

        With a self-loop:

        >>> graph = graphs.Graph([
        ...     [1, 3, 0],
        ...     [3, 0, 4],
        ...     [0, 0, 0],
        ... ])
        >>> graph.has_loops()
        True

        """
        return np.any(self.W.diagonal() != 0)

    def extract_components(self):
        r"""Split the graph into connected components.

        See :func:`is_connected` for the method used to determine
        connectedness.

        Returns
        -------
        graphs : list
            A list of graph structures. Each having its own node list and
            weight matrix. If the graph is directed, add into the info
            parameter the information about the source nodes and the sink
            nodes.

        Examples
        --------
        >>> from scipy import sparse
        >>> W = sparse.rand(10, 10, 0.2)
        >>> W = utils.symmetrize(W)
        >>> G = graphs.Graph(W)
        >>> components = G.extract_components()
        >>> has_sinks = 'sink' in components[0].info
        >>> sinks_0 = components[0].info['sink'] if has_sinks else []

        """
        if self.A.shape[0] != self.A.shape[1]:
            self.logger.error('Inconsistent shape to extract components. '
                              'Square matrix required.')
            return None

        if self.is_directed():
            raise NotImplementedError('Directed graphs not supported yet.')

        graphs = []

        visited = np.zeros(self.A.shape[0], dtype=bool)
        # indices = [] # Assigned but never used

        while not visited.all():
            # pick a node not visted yet
            stack = set(np.nonzero(~visited)[0][[0]])
            comp = []

            while len(stack):
                v = stack.pop()
                if not visited[v]:
                    comp.append(v)
                    visited[v] = True

                    # Add indices of nodes not visited yet and accessible from
                    # v
                    stack.update(set([idx for idx in self.A[v, :].nonzero()[1]
                                      if not visited[idx]]))

            comp = sorted(comp)
            self.logger.info(('Constructing subgraph for component of '
                              'size {}.').format(len(comp)))
            G = self.subgraph(comp)
            G.info = {'orig_idx': comp}
            graphs.append(G)

        return graphs

    def compute_laplacian(self, lap_type='combinatorial'):
        r"""Compute a graph Laplacian.

        For undirected graphs, the combinatorial Laplacian is defined as

        .. math:: L = D - W,

        where :math:`W` is the weighted adjacency matrix and :math:`D` the
        weighted degree matrix. The normalized Laplacian is defined as

        .. math:: L = I - D^{-1/2} W D^{-1/2},

        where :math:`I` is the identity matrix.

        For directed graphs, the Laplacians are built from a symmetrized
        version of the weighted adjacency matrix that is the average of the
        weighted adjacency matrix and its transpose. As the Laplacian is
        defined as the divergence of the gradient, it is not affected by the
        orientation of the edges.

        For both Laplacians, the diagonal entries corresponding to disconnected
        nodes (i.e., nodes with degree zero) are set to zero.

        Once computed, the Laplacian is accessible by the attribute :attr:`L`.

        Parameters
        ----------
        lap_type : {'combinatorial', 'normalized'}
            The kind of Laplacian to compute. Default is combinatorial.

        Examples
        --------

        Combinatorial and normalized Laplacians of an undirected graph.

        >>> graph = graphs.Graph([
        ...     [0, 2, 0],
        ...     [2, 0, 1],
        ...     [0, 1, 0],
        ... ])
        >>> graph.compute_laplacian('combinatorial')
        >>> graph.L.toarray()
        array([[ 2., -2.,  0.],
               [-2.,  3., -1.],
               [ 0., -1.,  1.]])
        >>> graph.compute_laplacian('normalized')
        >>> graph.L.toarray()
        array([[ 1.        , -0.81649658,  0.        ],
               [-0.81649658,  1.        , -0.57735027],
               [ 0.        , -0.57735027,  1.        ]])

        Combinatorial and normalized Laplacians of a directed graph.

        >>> graph = graphs.Graph([
        ...     [0, 2, 0],
        ...     [2, 0, 1],
        ...     [0, 0, 0],
        ... ])
        >>> graph.compute_laplacian('combinatorial')
        >>> graph.L.toarray()
        array([[ 2. , -2. ,  0. ],
               [-2. ,  2.5, -0.5],
               [ 0. , -0.5,  0.5]])
        >>> graph.compute_laplacian('normalized')
        >>> graph.L.toarray()
        array([[ 1.        , -0.89442719,  0.        ],
               [-0.89442719,  1.        , -0.4472136 ],
               [ 0.        , -0.4472136 ,  1.        ]])

        The Laplacian is defined as the divergence of the gradient.
        See :meth:`compute_differential_operator` for details.

        >>> graph = graphs.Path(20)
        >>> graph.compute_differential_operator()
        >>> L = graph.D.dot(graph.D.T)
        >>> np.all(L.toarray() == graph.L.toarray())
        True

        The Laplacians have a bounded spectrum.

        >>> G = graphs.Sensor(50)
        >>> G.compute_laplacian('combinatorial')
        >>> G.compute_fourier_basis()
        >>> -1e-10 < G.e[0] < 1e-10 < G.e[-1] < 2*np.max(G.dw)
        True
        >>> G.compute_laplacian('normalized')
        >>> G.compute_fourier_basis()
        >>> -1e-10 < G.e[0] < 1e-10 < G.e[-1] < 2
        True

        """

        if lap_type != self.lap_type:
            # Those attributes are invalidated when the Laplacian is changed.
            # Alternative: don't allow the user to change the Laplacian.
            self._lmax = None
            self._U = None
            self._e = None
            self._coherence = None
            self._D = None

        self.lap_type = lap_type

        if not self.is_directed():
            W = self.W
        else:
            W = utils.symmetrize(self.W, method='average')

        if lap_type == 'combinatorial':
            D = sparse.diags(self.dw)
            self.L = D - W
        elif lap_type == 'normalized':
            d = np.zeros(self.n_vertices)
            disconnected = (self.dw == 0)
            np.power(self.dw, -0.5, where=~disconnected, out=d)
            D = sparse.diags(d)
            self.L = sparse.identity(self.n_vertices) - D * W * D
            self.L[disconnected, disconnected] = 0
            self.L.eliminate_zeros()
        else:
            raise ValueError('Unknown Laplacian type {}'.format(lap_type))

    def _check_signal(self, s):
        r"""Check if signal is valid."""
        s = np.asanyarray(s)
        if s.shape[0] != self.n_vertices:
            raise ValueError('First dimension must be the number of vertices '
                             'G.N = {}, got {}.'.format(self.N, s.shape))
        return s

    def dirichlet_energy(self, x):
        r"""Compute the Dirichlet energy of a signal defined on the vertices.

        The Dirichlet energy of a signal :math:`x` is defined as

        .. math:: x^\top L x = \| \nabla_\mathcal{G} x \|_2^2
                             = \frac12 \sum_{i,j} W[i, j] (x[j] - x[i])^2

        for the combinatorial Laplacian, and

        .. math:: x^\top L x = \| \nabla_\mathcal{G} x \|_2^2
            = \frac12 \sum_{i,j} W[i, j]
              \left( \frac{x[j]}{d[j]} - \frac{x[i]}{d[i]} \right)^2

        for the normalized Laplacian, where :math:`d` is the weighted degree
        :attr:`dw`, :math:`\nabla_\mathcal{G} x = D^\top x` and :math:`D` is
        the differential operator :attr:`D`. See :meth:`grad` for the
        definition of the gradient :math:`\nabla_\mathcal{G}`.

        Parameters
        ----------
        x : array_like
            Signal of length :attr:`n_vertices` living on the vertices.

        Returns
        -------
        energy : float
            The Dirichlet energy of the graph signal.

        See Also
        --------
        grad : compute the gradient of a vertex signal

        Examples
        --------

        Non-directed graph:

        >>> graph = graphs.Path(5, directed=False)
        >>> signal = [0, 2, 2, 4, 4]
        >>> graph.dirichlet_energy(signal)
        8.0
        >>> # The Dirichlet energy is indeed the squared norm of the gradient.
        >>> graph.compute_differential_operator()
        >>> graph.grad(signal)
        array([2., 0., 2., 0.])

        Directed graph:

        >>> graph = graphs.Path(5, directed=True)
        >>> signal = [0, 2, 2, 4, 4]
        >>> graph.dirichlet_energy(signal)
        4.0
        >>> # The Dirichlet energy is indeed the squared norm of the gradient.
        >>> graph.compute_differential_operator()
        >>> graph.grad(signal)
        array([1.41421356, 0.        , 1.41421356, 0.        ])

        """
        x = self._check_signal(x)
        return x.T.dot(self.L.dot(x))

    @property
    def W(self):
        r"""Weighted adjacency matrix of the graph."""
        return self._adjacency

    @W.setter
    def W(self, value):
        # TODO: user can still do G.W[0, 0] = 1, or modify the passed W.
        raise AttributeError('In-place modification of the graph is not '
                             'supported. Create another Graph object.')

    @property
    def A(self):
        r"""Graph adjacency matrix (the binary version of W).

        The adjacency matrix defines which edges exist on the graph.
        It is represented as an N-by-N matrix of booleans.
        :math:`A_{i,j}` is True if :math:`W_{i,j} > 0`.
        """
        if self._A is None:
            self._A = self.W > 0
        return self._A

    @property
    def d(self):
        r"""The degree (number of neighbors) of vertices.

        For undirected graphs, the degree of a vertex is the number of vertices
        it is connected to.
        For directed graphs, the degree is the average of the in and out
        degrees, where the in degree is the number of incoming edges, and the
        out degree the number of outgoing edges.

        In both cases, the degree of the vertex :math:`v_i` is the average
        between the number of non-zero values in the :math:`i`-th column (the
        in degree) and the :math:`i`-th row (the out degree) of the weighted
        adjacency matrix :attr:`W`.

        Examples
        --------

        Undirected graph:

        >>> graph = graphs.Graph([
        ...     [0, 1, 0],
        ...     [1, 0, 2],
        ...     [0, 2, 0],
        ... ])
        >>> print(graph.d)  # Number of neighbors.
        [1 2 1]
        >>> print(graph.dw)  # Weighted degree.
        [1 3 2]

        Directed graph:

        >>> graph = graphs.Graph([
        ...     [0, 1, 0],
        ...     [0, 0, 2],
        ...     [0, 2, 0],
        ... ])
        >>> print(graph.d)  # Number of neighbors.
        [0.5 1.5 1. ]
        >>> print(graph.dw)  # Weighted degree.
        [0.5 2.5 2. ]

        """
        if self._d is None:
            if not self.is_directed():
                # Shortcut for undirected graphs.
                self._d = self.W.getnnz(axis=1)
                # axis=1 faster for CSR (https://stackoverflow.com/a/16391764)
            else:
                degree_in = self.W.getnnz(axis=0)
                degree_out = self.W.getnnz(axis=1)
                self._d = (degree_in + degree_out) / 2
        return self._d

    @property
    def dw(self):
        r"""The weighted degree of vertices.

        For undirected graphs, the weighted degree of the vertex :math:`v_i` is
        defined as

        .. math:: d[i] = \sum_j W[j, i] = \sum_j W[i, j],

        where :math:`W` is the weighted adjacency matrix :attr:`W`.

        For directed graphs, the weighted degree of the vertex :math:`v_i` is
        defined as

        .. math:: d[i] = \frac12 (d^\text{in}[i] + d^\text{out}[i])
                       = \frac12 (\sum_j W[j, i] + \sum_j W[i, j]),

        i.e., as the average of the in and out degrees.

        Examples
        --------

        Undirected graph:

        >>> graph = graphs.Graph([
        ...     [0, 1, 0],
        ...     [1, 0, 2],
        ...     [0, 2, 0],
        ... ])
        >>> print(graph.d)  # Number of neighbors.
        [1 2 1]
        >>> print(graph.dw)  # Weighted degree.
        [1 3 2]

        Directed graph:

        >>> graph = graphs.Graph([
        ...     [0, 1, 0],
        ...     [0, 0, 2],
        ...     [0, 2, 0],
        ... ])
        >>> print(graph.d)  # Number of neighbors.
        [0.5 1.5 1. ]
        >>> print(graph.dw)  # Weighted degree.
        [0.5 2.5 2. ]

        """
        if self._dw is None:
            if not self.is_directed():
                # Shortcut for undirected graphs.
                self._dw = np.ravel(self.W.sum(axis=0))
            else:
                degree_in = np.ravel(self.W.sum(axis=0))
                degree_out = np.ravel(self.W.sum(axis=1))
                self._dw = (degree_in + degree_out) / 2
        return self._dw

    @property
    def lmax(self):
        r"""Largest eigenvalue of the graph Laplacian.

        Can be exactly computed by :func:`compute_fourier_basis` or
        approximated by :func:`estimate_lmax`.
        """
        if self._lmax is None:
            self.logger.warning('The largest eigenvalue G.lmax is not '
                                'available, we need to estimate it. '
                                'Explicitly call G.estimate_lmax() or '
                                'G.compute_fourier_basis() '
                                'once beforehand to suppress the warning.')
            self.estimate_lmax()
        return self._lmax

    def estimate_lmax(self, method='lanczos'):
        r"""Estimate the Laplacian's largest eigenvalue (cached).

        The result is cached and accessible by the :attr:`lmax` property.

        Exact value given by the eigendecomposition of the Laplacian, see
        :func:`compute_fourier_basis`. That estimation is much faster than the
        eigendecomposition.

        Parameters
        ----------
        method : {'lanczos', 'bounds'}
            Whether to estimate the largest eigenvalue with the implicitly
            restarted Lanczos method, or to return an upper bound on the
            spectrum of the Laplacian.

        Notes
        -----
        Runs the implicitly restarted Lanczos method (as implemented in
        :func:`scipy.sparse.linalg.eigsh`) with a large tolerance, then
        increases the calculated largest eigenvalue by 1 percent. For much of
        the PyGSP2 machinery, we need to approximate filter kernels on an
        interval that contains the spectrum of L. The only cost of using a
        larger interval is that the polynomial approximation over the larger
        interval may be a slightly worse approximation on the actual spectrum.
        As this is a very mild effect, it is not necessary to obtain very tight
        bounds on the spectrum of L.

        A faster but less tight alternative is to use known algebraic bounds on
        the graph Laplacian.

        Examples
        --------
        >>> G = graphs.Logo()
        >>> G.compute_fourier_basis()  # True value.
        >>> print('{:.2f}'.format(G.lmax))
        13.78
        >>> G.estimate_lmax(method='lanczos')  # Estimate.
        >>> print('{:.2f}'.format(G.lmax))
        13.92
        >>> G.estimate_lmax(method='bounds')  # Upper bound.
        >>> print('{:.2f}'.format(G.lmax))
        18.58

        """
        if method == self._lmax_method:
            return
        self._lmax_method = method

        if method == 'lanczos':
            try:
                # We need to cast the matrix L to a supported type.
                # TODO: not good for memory. Cast earlier?
                lmax = sparse.linalg.eigsh(self.L.asfptype(), k=1, tol=5e-3,
                                           ncv=min(self.N, 10),
                                           return_eigenvectors=False)
                lmax = lmax[0]
                assert lmax <= self._get_upper_bound() + 1e-12
                lmax *= 1.01  # Increase by 1% to be robust to errors.
                self._lmax = lmax
            except sparse.linalg.ArpackNoConvergence:
                raise ValueError('The Lanczos method did not converge. '
                                 'Try to use bounds.')

        elif method == 'bounds':
            self._lmax = self._get_upper_bound()

        else:
            raise ValueError('Unknown method {}'.format(method))

    def _get_upper_bound(self):
        r"""Return an upper bound on the eigenvalues of the Laplacian."""

        if self.lap_type == 'normalized':
            return 2  # Equal iff the graph is bipartite.
        elif self.lap_type == 'combinatorial':
            bounds = []
            # Equal for full graphs.
            bounds += [self.n_vertices * np.max(self.W)]
            # Gershgorin circle theorem. Equal for regular bipartite graphs.
            # Special case of the below bound.
            bounds += [2 * np.max(self.dw)]
            # Anderson, Morley, Eigenvalues of the Laplacian of a graph.
            # Equal for regular bipartite graphs.
            if self.n_edges > 0:
                sources, targets, _ = self.get_edge_list()
                bounds += [np.max(self.dw[sources] + self.dw[targets])]
            # Merris, A note on Laplacian graph eigenvalues.
            if not self.is_directed():
                W = self.W
            else:
                W = utils.symmetrize(self.W, method='average')
            m = W.dot(self.dw) / self.dw  # Mean degree of adjacent vertices.
            bounds += [np.max(self.dw + m)]
            # Good review: On upper bounds for Laplacian graph eigenvalues.
            return min(bounds)
        else:
            raise ValueError('Unknown Laplacian type '
                             '{}'.format(self.lap_type))

    def get_edge_list(self):
        r"""Return an edge list, an alternative representation of the graph.

        Each edge :math:`e_k = (v_i, v_j) \in \mathcal{E}` from :math:`v_i` to
        :math:`v_j` is associated with the weight :math:`W[i, j]`. For each
        edge :math:`e_k`, the method returns :math:`(i, j, W[i, j])` as
        `(sources[k], targets[k], weights[k])`, with :math:`i \in [0,
        |\mathcal{V}|-1], j \in [0, |\mathcal{V}|-1], k \in [0,
        |\mathcal{E}|-1]`.

        Returns
        -------
        sources : vector of int
            Source node indices.
        targets : vector of int
            Target node indices.
        weights : vector of float
            Edge weights.

        Notes
        -----
        The weighted adjacency matrix is the canonical form used in this
        package to represent a graph as it is the easiest to work with when
        considering spectral methods.

        Edge orientation (i.e., which node is the source or the target) is
        arbitrary for undirected graphs.
        The implementation uses the upper triangular part of the adjacency
        matrix, hence :math:`i \leq j \ \forall k`.

        Examples
        --------

        Edge list of a directed graph.

        >>> graph = graphs.Graph([
        ...     [0, 3, 0],
        ...     [3, 0, 4],
        ...     [0, 0, 0],
        ... ])
        >>> sources, targets, weights = graph.get_edge_list()
        >>> list(sources), list(targets), list(weights)
        ([0, 1, 1], [1, 0, 2], [3, 3, 4])

        Edge list of an undirected graph.

        >>> graph = graphs.Graph([
        ...     [0, 3, 0],
        ...     [3, 0, 4],
        ...     [0, 4, 0],
        ... ])
        >>> sources, targets, weights = graph.get_edge_list()
        >>> list(sources), list(targets), list(weights)
        ([0, 1], [1, 2], [3, 4])

        """

        if self.is_directed():
            W = self.W.tocoo()
        else:
            W = sparse.triu(self.W, format='coo')

        sources = W.row
        targets = W.col
        weights = W.data

        assert self.n_edges == sources.size == targets.size == weights.size
        return sources, targets, weights

    def plot(self, vertex_color=None, vertex_size=None, highlight=[],
             edges=None, edge_color=None, edge_width=None,
             indices=False, colorbar=True, limits=None, ax=None,
             title=None, backend=None, cmap=None, alphan=1, alphav=1, edge_weights=None):
        r"""Docstring overloaded at import time."""
        from pygsp2.plotting import _plot_graph
        return _plot_graph(self, vertex_color=vertex_color,
                           vertex_size=vertex_size, highlight=highlight,
                           edges=edges, indices=indices, colorbar=colorbar,
                           edge_color=edge_color, edge_width=edge_width,
                           limits=limits, ax=ax, title=title, backend=backend, 
                           cmap=cmap, alphan=alphan, alphav=alphav, edge_weights=edge_weights)

    def plot_signal(self, *args, **kwargs):
        r"""Deprecated, use plot() instead."""
        return self.plot(*args, **kwargs)

    def plot_spectrogram(self, node_idx=None):
        r"""Docstring overloaded at import time."""
        from pygsp2.plotting import _plot_spectrogram
        _plot_spectrogram(self, node_idx=node_idx)



# -*- coding: utf-8 -*-

import os

import numpy as np


def _import_networkx():
    try:
        import networkx as nx
    except Exception as e:
        raise ImportError('Cannot import networkx. Use graph-tool or try to '
                          'install it with pip (or conda) install networkx. '
                          'Original exception: {}'.format(e))
    return nx


def _import_graphtool():
    try:
        import graph_tool as gt
    except Exception as e:
        raise ImportError('Cannot import graph-tool. Use networkx or try to '
                          'install it. Original exception: {}'.format(e))
    return gt


class IOMixIn(object):

    def _break_signals(self):
        r"""Break N-dimensional signals into N 1D signals."""
        for name in list(self.signals.keys()):
            if self.signals[name].ndim == 2:
                for i, signal_1d in enumerate(self.signals[name].T):
                    self.signals[name + '_' + str(i)] = signal_1d
                del self.signals[name]

    def _join_signals(self):
        r"""Join N 1D signals into one N-dimensional signal."""
        joined = dict()
        for name in self.signals:
            name_base = name.rsplit('_', 1)[0]
            names = joined.get(name_base, list())
            names.append(name)
            joined[name_base] = names
        for name_base, names in joined.items():
            if len(names) > 1:
                names = sorted(names)  # ensure dim ordering (_0, _1, etc.)
                signal_nd = np.stack([self.signals[n] for n in names], axis=1)
                self.signals[name_base] = signal_nd
                for name in names:
                    del self.signals[name]

    def to_networkx(self):
        r"""Export the graph to NetworkX.

        Edge weights are stored as an edge attribute,
        under the name "weight".

        Signals are stored as node attributes,
        under their name in the :attr:`signals` dictionary.
        `N`-dimensional signals are broken into `N` 1-dimensional signals.
        They will eventually be joined back together on import.

        Returns
        -------
        graph : :class:`networkx.Graph`
            A NetworkX graph object.

        See Also
        --------
        to_graphtool : export to graph-tool
        save : save to a file

        Examples
        --------
        >>> import networkx as nx
        >>> from matplotlib import pyplot as plt
        >>> graph = graphs.Path(4, directed=True)
        >>> graph.set_signal(np.full(4, 2.3), 'signal')
        >>> graph = graph.to_networkx()
        >>> print(nx.info(graph))
        DiGraph named 'Path' with 4 nodes and 3 edges
        >>> nx.is_directed(graph)
        True
        >>> graph.nodes()
        NodeView((0, 1, 2, 3))
        >>> graph.edges()
        OutEdgeView([(0, 1), (1, 2), (2, 3)])
        >>> graph.nodes()[2]
        {'signal': 2.3}
        >>> graph.edges()[(0, 1)]
        {'weight': 1.0}
        >>> # nx.draw(graph, with_labels=True)

        Another common goal is to use NetworkX to compute some properties to be
        be imported back in the PyGSP2 as signals.

        >>> import networkx as nx
        >>> from matplotlib import pyplot as plt
        >>> graph = graphs.Sensor(100, seed=42)
        >>> graph.set_signal(graph.coords, 'coords')
        >>> graph = graph.to_networkx()
        >>> betweenness = nx.betweenness_centrality(graph, weight='weight')
        >>> nx.set_node_attributes(graph, betweenness, 'betweenness')
        >>> graph = graphs.Graph.from_networkx(graph)
        >>> graph.compute_fourier_basis()
        >>> graph.set_coordinates(graph.signals['coords'])
        >>> fig, axes = plt.subplots(1, 2)
        >>> _ = graph.plot(graph.signals['betweenness'], ax=axes[0])
        >>> _ = axes[1].plot(graph.e, graph.gft(graph.signals['betweenness']))

        """
        nx = _import_networkx()

        def convert(number):
            # NetworkX accepts arbitrary python objects as attributes, but:
            # * the GEXF writer does not accept any NumPy types (on signals),
            # * the GraphML writer does not accept NumPy ints.
            if issubclass(number.dtype.type, (np.integer, np.bool_)):
                return int(number)
            else:
                return float(number)

        def edges():
            for source, target, weight in zip(*self.get_edge_list()):
                yield int(source), int(target), {'weight': convert(weight)}

        def nodes():
            for vertex in range(self.n_vertices):
                signals = {name: convert(signal[vertex])
                           for name, signal in self.signals.items()}
                yield vertex, signals

        self._break_signals()
        graph = nx.DiGraph() if self.is_directed() else nx.Graph()
        graph.add_nodes_from(nodes())
        graph.add_edges_from(edges())
        graph.name = self.__class__.__name__
        return graph

    def to_graphtool(self):
        r"""Export the graph to graph-tool.

        Edge weights are stored as an edge property map,
        under the name "weight".

        Signals are stored as vertex property maps,
        under their name in the :attr:`signals` dictionary.
        `N`-dimensional signals are broken into `N` 1-dimensional signals.
        They will eventually be joined back together on import.

        Returns
        -------
        graph : :class:`graph_tool.Graph`
            A graph-tool graph object.

        See Also
        --------
        to_networkx : export to NetworkX
        save : save to a file

        Examples
        --------
        >>> import graph_tool as gt
        >>> import graph_tool.draw
        >>> from matplotlib import pyplot as plt
        >>> graph = graphs.Path(4, directed=True)
        >>> graph.set_signal(np.full(4, 2.3), 'signal')
        >>> graph = graph.to_graphtool()
        >>> graph.is_directed()
        True
        >>> graph.vertex_properties['signal'][2]
        2.3
        >>> graph.edge_properties['weight'][graph.edge(0, 1)]
        1.0
        >>> # gt.draw.graph_draw(graph, vertex_text=graph.vertex_index)

        Another common goal is to use graph-tool to compute some properties to
        be imported back in the PyGSP2 as signals.

        >>> import graph_tool as gt
        >>> import graph_tool.centrality
        >>> from matplotlib import pyplot as plt
        >>> graph = graphs.Sensor(100, seed=42)
        >>> graph.set_signal(graph.coords, 'coords')
        >>> graph = graph.to_graphtool()
        >>> vprop, eprop = gt.centrality.betweenness(
        ...     graph, weight=graph.edge_properties['weight'])
        >>> graph.vertex_properties['betweenness'] = vprop
        >>> graph = graphs.Graph.from_graphtool(graph)
        >>> graph.compute_fourier_basis()
        >>> graph.set_coordinates(graph.signals['coords'])
        >>> fig, axes = plt.subplots(1, 2)
        >>> _ = graph.plot(graph.signals['betweenness'], ax=axes[0])
        >>> _ = axes[1].plot(graph.e, graph.gft(graph.signals['betweenness']))

        """

        gt = _import_graphtool()
        graph = gt.Graph(directed=self.is_directed())

        sources, targets, weights = self.get_edge_list()
        graph.add_edge_list(zip(sources, targets))
        prop = graph.new_edge_property(gt._gt_type(weights.dtype))
        prop.get_array()[:] = weights
        graph.edge_properties['weight'] = prop

        self._break_signals()
        for name, signal in self.signals.items():
            prop = graph.new_vertex_property(gt._gt_type(signal.dtype))
            prop.get_array()[:] = signal
            graph.vertex_properties[name] = prop

        return graph

    @classmethod
    def from_networkx(cls, graph, weight='weight'):
        r"""Import a graph from NetworkX.

        Edge weights are retrieved as an edge attribute,
        under the name specified by the ``weight`` parameter.

        Signals are retrieved from node attributes,
        and stored in the :attr:`signals` dictionary under the attribute name.
        `N`-dimensional signals that were broken during export are joined.

        Parameters
        ----------
        graph : :class:`networkx.Graph`
            A NetworkX graph object.
        weight : string or None, optional
            The edge attribute that holds the numerical values used as the edge
            weights. All edge weights are set to 1 if None, or not found.

        Returns
        -------
        graph : :class:`~pygsp2.graphs.Graph`
            A PyGSP2 graph object.

        Notes
        -----

        The nodes are ordered according to :meth:`networkx.Graph.nodes`.

        In NetworkX, node attributes need not be set for every node.
        If a node attribute is not set for a node, a NaN is assigned to the
        corresponding signal for that node.

        If the graph is a :class:`networkx.MultiGraph`, multiedges are
        aggregated by summation.

        See Also
        --------
        from_graphtool : import from graph-tool
        load : load from a file

        Examples
        --------
        >>> import networkx as nx
        >>> graph = nx.Graph()
        >>> graph.add_edge(1, 2, weight=0.2)
        >>> graph.add_edge(2, 3, weight=0.9)
        >>> graph.add_node(4, sig=3.1416)
        >>> graph.nodes()
        NodeView((1, 2, 3, 4))
        >>> graph = graphs.Graph.from_networkx(graph)
        >>> graph.W.toarray()
        array([[0. , 0.2, 0. , 0. ],
               [0.2, 0. , 0.9, 0. ],
               [0. , 0.9, 0. , 0. ],
               [0. , 0. , 0. , 0. ]])
        >>> graph.signals
        {'sig': array([   nan,    nan,    nan, 3.1416])}

        """
        nx = _import_networkx()
        from .graph import Graph

        adjacency = nx.to_scipy_sparse_matrix(graph, weight=weight)
        graph_pg = Graph(adjacency)

        for i, node in enumerate(graph.nodes()):
            for name in graph.nodes[node].keys():
                try:
                    signal = graph_pg.signals[name]
                except KeyError:
                    signal = np.full(graph_pg.n_vertices, np.nan)
                    graph_pg.set_signal(signal, name)
                try:
                    signal[i] = graph.nodes[node][name]
                except KeyError:
                    pass  # attribute not set for node

        graph_pg._join_signals()
        return graph_pg

    @classmethod
    def from_graphtool(cls, graph, weight='weight'):
        r"""Import a graph from graph-tool.

        Edge weights are retrieved as an edge property,
        under the name specified by the ``weight`` parameter.

        Signals are retrieved from node properties,
        and stored in the :attr:`signals` dictionary under the property name.
        `N`-dimensional signals that were broken during export are joined.

        Parameters
        ----------
        graph : :class:`graph_tool.Graph`
            A graph-tool graph object.
        weight : string
            The edge property that holds the numerical values used as the edge
            weights. All edge weights are set to 1 if None, or not found.

        Returns
        -------
        graph : :class:`~pygsp2.graphs.Graph`
            A PyGSP2 graph object.

        Notes
        -----

        If the graph has multiple edge connecting the same two nodes, a sum
        over the edges is taken to merge them.

        See Also
        --------
        from_networkx : import from NetworkX
        load : load from a file

        Examples
        --------
        >>> import graph_tool as gt
        >>> graph = gt.Graph(directed=False)
        >>> e1 = graph.add_edge(0, 1)
        >>> e2 = graph.add_edge(1, 2)
        >>> v = graph.add_vertex()
        >>> eprop = graph.new_edge_property("double")
        >>> eprop[e1] = 0.2
        >>> eprop[graph.edge(1, 2)] = 0.9
        >>> graph.edge_properties["weight"] = eprop
        >>> vprop = graph.new_vertex_property("double", val=np.nan)
        >>> vprop[3] = 3.1416
        >>> graph.vertex_properties["sig"] = vprop
        >>> graph = graphs.Graph.from_graphtool(graph)
        >>> graph.W.toarray()
        array([[0. , 0.2, 0. , 0. ],
               [0.2, 0. , 0.9, 0. ],
               [0. , 0.9, 0. , 0. ],
               [0. , 0. , 0. , 0. ]])
        >>> graph.signals
        {'sig': PropertyArray([   nan,    nan,    nan, 3.1416])}

        """
        gt = _import_graphtool()
        import graph_tool.spectral
        from .graph import Graph

        weight = graph.edge_properties.get(weight, None)
        adjacency = gt.spectral.adjacency(graph, weight=weight)
        graph_pg = Graph(adjacency.T)

        for name, signal in graph.vertex_properties.items():
            graph_pg.set_signal(signal.get_array(), name)

        graph_pg._join_signals()
        return graph_pg

    @classmethod
    def load(cls, path, fmt=None, backend=None):
        r"""Load a graph from a file.

        Edge weights are retrieved as an edge attribute named "weight".

        Signals are retrieved from node attributes,
        and stored in the :attr:`signals` dictionary under the attribute name.
        `N`-dimensional signals that were broken during export are joined.

        Parameters
        ----------
        path : string
            Path to the file from which to load the graph.
        fmt : {'graphml', 'gml', 'gexf', None}, optional
            Format in which the graph is saved.
            Guessed from the filename extension if None.
        backend : {'networkx', 'graph-tool', None}, optional
            Library used to load the graph. Automatically chosen if None.

        Returns
        -------
        graph : :class:`Graph`
            The loaded graph.

        See Also
        --------
        save : save a graph to a file
        from_networkx : load with NetworkX then import in the PyGSP2
        from_graphtool : load with graph-tool then import in the PyGSP2

        Notes
        -----

        A lossless round-trip is only guaranteed if the graph (and its signals)
        is saved and loaded with the same backend.

        Loading from other formats is possible by loading in NetworkX or
        graph-tool, and importing to the PyGSP2.
        The proposed formats are however tested for faithful round-trips.

        Examples
        --------
        >>> graph = graphs.Logo()
        >>> graph.save('logo.graphml')
        >>> graph = graphs.Graph.load('logo.graphml')
        >>> import os
        >>> os.remove('logo.graphml')

        """

        if fmt is None:
            fmt = os.path.splitext(path)[1][1:]
        if fmt not in ['graphml', 'gml', 'gexf']:
            raise ValueError('Unsupported format {}.'.format(fmt))

        def load_networkx(path, fmt):
            nx = _import_networkx()
            load = getattr(nx, 'read_' + fmt)
            graph = load(path)
            return cls.from_networkx(graph)

        def load_graphtool(path, fmt):
            gt = _import_graphtool()
            graph = gt.load_graph(path, fmt=fmt)
            return cls.from_graphtool(graph)

        if backend == 'networkx':
            return load_networkx(path, fmt)
        elif backend == 'graph-tool':
            return load_graphtool(path, fmt)
        elif backend is None:
            try:
                return load_networkx(path, fmt)
            except ImportError:
                try:
                    return load_graphtool(path, fmt)
                except ImportError:
                    raise ImportError('Cannot import networkx nor graph-tool.')
        else:
            raise ValueError('Unknown backend {}.'.format(backend))

    def save(self, path, fmt=None, backend=None):
        r"""Save the graph to a file.

        Edge weights are stored as an edge attribute,
        under the name "weight".

        Signals are stored as node attributes,
        under their name in the :attr:`signals` dictionary.
        `N`-dimensional signals are broken into `N` 1-dimensional signals.
        They will eventually be joined back together on import.

        Supported formats are:

        * GraphML_, a comprehensive XML format.
          Supported by NetworkX_, graph-tool_, NetworKit_, igraph_, Gephi_,
          Cytoscape_, SocNetV_.
        * GML_ (Graph Modelling Language), a simple non-XML format.
          Supported by NetworkX_, graph-tool_, NetworKit_, igraph_, Gephi_,
          Cytoscape_, SocNetV_, Tulip_.
        * GEXF_ (Graph Exchange XML Format), Gephi's XML format.
          Supported by NetworkX_, NetworKit_, Gephi_, Tulip_, ngraph_.

        If unsure, we recommend GraphML_.

        .. _GraphML: https://en.wikipedia.org/wiki/GraphML
        .. _GML: https://en.wikipedia.org/wiki/Graph_Modelling_Language
        .. _GEXF: https://gephi.org/gexf
        .. _NetworkX: https://networkx.org
        .. _graph-tool: https://graph-tool.skewed.de
        .. _NetworKit: https://networkit.github.io
        .. _igraph: https://igraph.org
        .. _ngraph: https://github.com/anvaka/ngraph
        .. _Gephi: https://gephi.org
        .. _Cytoscape: https://cytoscape.org
        .. _SocNetV: https://socnetv.org
        .. _Tulip: https://tulip.labri.fr

        Parameters
        ----------
        path : string
            Path to the file where the graph is to be saved.
        fmt : {'graphml', 'gml', 'gexf', None}, optional
            Format in which to save the graph.
            Guessed from the filename extension if None.
        backend : {'networkx', 'graph-tool', None}, optional
            Library used to load the graph. Automatically chosen if None.

        See Also
        --------
        load : load a graph from a file
        to_networkx : export as a NetworkX graph, and save with NetworkX
        to_graphtool : export as a graph-tool graph, and save with graph-tool

        Notes
        -----

        A lossless round-trip is only guaranteed if the graph (and its signals)
        is saved and loaded with the same backend.

        Saving in other formats is possible by exporting to NetworkX or
        graph-tool, and using their respective saving functionality.
        The proposed formats are however tested for faithful round-trips.

        Edge weights and signal values are rounded at the sixth decimal when
        saving in ``fmt='gml'`` with ``backend='graph-tool'``.

        Examples
        --------
        >>> graph = graphs.Logo()
        >>> graph.save('logo.graphml')
        >>> graph = graphs.Graph.load('logo.graphml')
        >>> import os
        >>> os.remove('logo.graphml')

        """

        if fmt is None:
            fmt = os.path.splitext(path)[1][1:]
        if fmt not in ['graphml', 'gml', 'gexf']:
            raise ValueError('Unsupported format {}.'.format(fmt))

        def save_networkx(graph, path, fmt):
            nx = _import_networkx()
            graph = graph.to_networkx()
            save = getattr(nx, 'write_' + fmt)
            save(graph, path)

        def save_graphtool(graph, path, fmt):
            graph = graph.to_graphtool()
            graph.save(path, fmt=fmt)

        if backend == 'networkx':
            save_networkx(self, path, fmt)
        elif backend == 'graph-tool':
            save_graphtool(self, path, fmt)
        elif backend is None:
            try:
                save_networkx(self, path, fmt)
            except ImportError:
                try:
                    save_graphtool(self, path, fmt)
                except ImportError:
                    raise ImportError('Cannot import networkx nor graph-tool.')
        else:
            raise ValueError('Unknown backend {}.'.format(backend))



# -*- coding: utf-8 -*-

r"""
The :mod:`pygsp2.utils` module implements some utility functions used throughout
the package.
"""

from __future__ import division

import sys
import logging
import functools
import pkgutil
import io

import numpy as np
from scipy import sparse
import scipy.io


def build_logger(name):
    logger = logging.getLogger(name)

    if not logger.handlers:
        formatter = logging.Formatter(
            "%(asctime)s:[%(levelname)s](%(name)s.%(funcName)s): %(message)s")

        steam_handler = logging.StreamHandler()
        steam_handler.setLevel(logging.DEBUG)
        steam_handler.setFormatter(formatter)

        logger.setLevel(logging.DEBUG)
        logger.addHandler(steam_handler)

    return logger


logger = build_logger(__name__)


def filterbank_handler(func):

    # Preserve documentation of func.
    @functools.wraps(func)
    def inner(f, *args, **kwargs):

        if 'i' in kwargs:
            return func(f, *args, **kwargs)

        elif f.Nf <= 1:
            return func(f, *args, **kwargs)

        else:
            output = []
            for i in range(f.Nf):
                output.append(func(f, *args, i=i, **kwargs))
            return output

    return inner


def loadmat(path):
    r"""
    Load a matlab data file.

    Parameters
    ----------
    path : string
        Path to the mat file from the data folder, without the .mat extension.

    Returns
    -------
    data : dict
        dictionary with variable names as keys, and loaded matrices as
        values.

    Examples
    --------
    >>> from pygsp2 import utils
    >>> data = utils.loadmat('pointclouds/bunny')
    >>> data['bunny'].shape
    (2503, 3)

    """
    data = pkgutil.get_data('pygsp2', 'data/' + path + '.mat')
    data = io.BytesIO(data)
    return scipy.io.loadmat(data)


def distanz(x, y=None):
    r"""
    Calculate the distance between two colon vectors.

    Parameters
    ----------
    x : ndarray
        First colon vector
    y : ndarray
        Second colon vector

    Returns
    -------
    d : ndarray
        Distance between x and y

    Examples
    --------
    >>> from pygsp2 import utils
    >>> x = np.arange(3)
    >>> utils.distanz(x, x)
    array([[0., 1., 2.],
           [1., 0., 1.],
           [2., 1., 0.]])

    """
    try:
        x.shape[1]
    except IndexError:
        x = x.reshape(1, x.shape[0])

    if y is None:
        y = x

    else:
        try:
            y.shape[1]
        except IndexError:
            y = y.reshape(1, y.shape[0])

    rx, cx = x.shape
    ry, cy = y.shape

    # Size verification
    if rx != ry:
        raise ValueError("The sizes of x and y do not fit")

    xx = (x * x).sum(axis=0)
    yy = (y * y).sum(axis=0)
    xy = np.dot(x.T, y)

    d = abs(np.kron(np.ones((cy, 1)), xx).T +
            np.kron(np.ones((cx, 1)), yy) - 2 * xy)

    return np.sqrt(d)


def resistance_distance(G):
    r"""
    Compute the resistance distances of a graph.

    Parameters
    ----------
    G : Graph or sparse matrix
        Graph structure or Laplacian matrix (L)

    Returns
    -------
    rd : sparse matrix
        distance matrix

    References
    ----------
    :cite:`klein1993resistance`
    """

    if sparse.issparse(G):
        L = G.tocsc()

    else:
        if G.lap_type != 'combinatorial':
            raise ValueError('Need a combinatorial Laplacian.')
        L = G.L.tocsc()

    try:
        pseudo = sparse.linalg.inv(L)
    except RuntimeError:
        pseudo = sparse.lil_matrix(np.linalg.pinv(L.toarray()))

    N = np.shape(L)[0]
    d = sparse.csc_matrix(pseudo.diagonal())
    rd = sparse.kron(d, sparse.csc_matrix(np.ones((N, 1)))).T \
        + sparse.kron(d, sparse.csc_matrix(np.ones((N, 1)))) \
        - pseudo - pseudo.T

    return rd


def symmetrize(W, method='average'):
    r"""
    Symmetrize a square matrix.

    Parameters
    ----------
    W : array_like
        Square matrix to be symmetrized
    method : string
        * 'average' : symmetrize by averaging with the transpose. Most useful
          when transforming a directed graph to an undirected one.
        * 'maximum' : symmetrize by taking the maximum with the transpose.
          Similar to 'fill' except that ambiguous entries are resolved by
          taking the largest value.
        * 'fill' : symmetrize by filling in the zeros in both the upper and
          lower triangular parts. Ambiguous entries are resolved by averaging
          the values.
        * 'tril' : symmetrize by considering the lower triangular part only.
        * 'triu' : symmetrize by considering the upper triangular part only.

    Notes
    -----
    You can have the sum by multiplying the average by two. It is however not a
    good candidate for this function as it modifies an already symmetric
    matrix.

    Examples
    --------
    >>> from pygsp2 import utils
    >>> W = np.array([[0, 3, 0], [3, 1, 6], [4, 2, 3]], dtype=float)
    >>> W
    array([[0., 3., 0.],
           [3., 1., 6.],
           [4., 2., 3.]])
    >>> utils.symmetrize(W, method='average')
    array([[0., 3., 2.],
           [3., 1., 4.],
           [2., 4., 3.]])
    >>> 2 * utils.symmetrize(W, method='average')
    array([[0., 6., 4.],
           [6., 2., 8.],
           [4., 8., 6.]])
    >>> utils.symmetrize(W, method='maximum')
    array([[0., 3., 4.],
           [3., 1., 6.],
           [4., 6., 3.]])
    >>> utils.symmetrize(W, method='fill')
    array([[0., 3., 4.],
           [3., 1., 4.],
           [4., 4., 3.]])
    >>> utils.symmetrize(W, method='tril')
    array([[0., 3., 4.],
           [3., 1., 2.],
           [4., 2., 3.]])
    >>> utils.symmetrize(W, method='triu')
    array([[0., 3., 0.],
           [3., 1., 6.],
           [0., 6., 3.]])

    """
    if W.shape[0] != W.shape[1]:
        raise ValueError('Matrix must be square.')

    if method == 'average':
        return (W + W.T) / 2

    elif method == 'maximum':
        if sparse.issparse(W):
            bigger = (W.T > W)
            return W - W.multiply(bigger) + W.T.multiply(bigger)
        else:
            return np.maximum(W, W.T)

    elif method == 'fill':
        A = (W > 0)  # Boolean type.
        if sparse.issparse(W):
            mask = (A + A.T) - A
            W = W + mask.multiply(W.T)
        else:
            # Numpy boolean subtract is deprecated.
            mask = np.logical_xor(np.logical_or(A, A.T), A)
            W = W + mask * W.T
        return symmetrize(W, method='average')  # Resolve ambiguous entries.

    elif method in ['tril', 'triu']:
        if sparse.issparse(W):
            tri = getattr(sparse, method)
        else:
            tri = getattr(np, method)
        W = tri(W)
        return symmetrize(W, method='maximum')

    else:
        raise ValueError('Unknown symmetrization method {}.'.format(method))


def rescale_center(x):
    r"""
    Rescale and center data, e.g. embedding coordinates.

    Parameters
    ----------
    x : ndarray
        Data to be rescaled.

    Returns
    -------
    r : ndarray
        Rescaled data.

    Examples
    --------
    >>> from pygsp2 import utils
    >>> x = np.array([[1, 6], [2, 5], [3, 4]])
    >>> utils.rescale_center(x)
    array([[-1. ,  1. ],
           [-0.6,  0.6],
           [-0.2,  0.2]])

    """
    N = x.shape[1]
    y = x - np.kron(np.ones((1, N)), np.mean(x, axis=1)[:, np.newaxis])
    c = np.amax(y)
    r = y / c

    return r


def compute_log_scales(lmin, lmax, Nscales, t1=1, t2=2):
    r"""
    Compute logarithm scales for wavelets.

    Parameters
    ----------
    lmin : float
        Smallest non-zero eigenvalue.
    lmax : float
        Largest eigenvalue, i.e. :py:attr:`pygsp2.graphs.Graph.lmax`.
    Nscales : int
        Number of scales.

    Returns
    -------
    scales : ndarray
        List of scales of length Nscales.

    Examples
    --------
    >>> from pygsp2 import utils
    >>> utils.compute_log_scales(1, 10, 3)
    array([2.       , 0.4472136, 0.1      ])

    """
    scale_min = t1 / lmax
    scale_max = t2 / lmin
    return np.exp(np.linspace(np.log(scale_max), np.log(scale_min), Nscales))


def to_sparse(i, j, v, m, n):
    """
    Create and compressing a matrix that have many zeros
    Parameters:
        i: 1-D array representing the index 1 values 
            Size n1
        j: 1-D array representing the index 2 values 
            Size n1
        v: 1-D array representing the values 
            Size n1
        m: integer representing x size of the matrix >= n1
        n: integer representing y size of the matrix >= n1
    Returns:
        s: 2-D array
            Matrix full of zeros excepting values v at indexes i, j
    """
    return sparse.csr_matrix((v, (i, j)), shape=(m, n))


def sum_squareform(n):
    """ Returns sparse matrix that sums the squareform of a vector. Reference
    from the unlocbox toolbox function for matlab.

    Parameters
    ----------
    n: int, number of nodes in the spare matrix.

    Returns
    -------
    S:    matrix so that S*w = sum(W) for vector w = squareform(W)
    St:   the adjoint of S

    Reference:
    https://epfl-lts2.github.io/gspbox-html/doc/learn_graph/gsp_learn_graph_log_degrees.html

    """
    # number of columns is the length of w given size of W
    ncols = int((n-1)*(n)/2)

    I = np.zeros([ncols])
    J = np.zeros([ncols])

    # offset
    k = 0
    for i in np.arange(1, n):
        I[k: k + (n-i)] = np.arange(i, n)
        k = k + (n-i)

    k = 0
    for i in np.arange(1, n):
        J[k: k + (n-i)] = i-1
        k = k + (n-i)

    i = np.array(np.hstack([np.arange(0, ncols), np.arange(0, ncols)]))
    j = np.hstack([I, J]).squeeze().T.ravel()
    s = np.ones(len(i))
    m = ncols

    St = to_sparse(i, j, s, m, n)

    S = St.T

    return (S, St)



# -*- coding: utf-8 -*-

"""
Test suite for the plotting module of the pygsp2 package.

"""

import unittest
import os

import numpy as np
from matplotlib import pyplot as plt
from skimage import data, img_as_float

from pygsp2 import graphs, filters, plotting


class TestGraphs(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls._img = img_as_float(data.camera()[::16, ::16])

    def tearDown(cls):
        plotting.close_all()

    def test_all_graphs(self):
        r"""
        Plot all graphs which have coordinates.
        With and without signal.
        With both backends.
        """

        # Graphs who are not embedded, i.e., have no coordinates.
        COORDS_NO = {
            'Graph',
            'BarabasiAlbert',
            'ErdosRenyi',
            'FullConnected',
            'RandomRegular',
            'StochasticBlockModel',
            }

        Gs = []
        for classname in dir(graphs):

            if not classname[0].isupper():
                # Not a Graph class but a submodule or private stuff.
                continue
            elif classname in COORDS_NO:
                continue
            elif classname == 'ImgPatches':
                # Coordinates are not in 2D or 3D.
                continue

            Graph = getattr(graphs, classname)

            # Classes who require parameters.
            if classname == 'NNGraph':
                Xin = np.arange(90).reshape(30, 3)
                Gs.append(Graph(Xin))
            elif classname == 'Grid2dImgPatches':
                Gs.append(Graph(img=self._img, patch_shape=(3, 3)))
            elif classname == 'LineGraph':
                Gs.append(Graph(graphs.Sensor(20, seed=42)))
            else:
                Gs.append(Graph())

            # Add more test cases.
            if classname == 'TwoMoons':
                Gs.append(Graph(moontype='standard'))
                Gs.append(Graph(moontype='synthesized'))
            elif classname == 'Cube':
                Gs.append(Graph(nb_dim=2))
                Gs.append(Graph(nb_dim=3))
            elif classname == 'DavidSensorNet':
                Gs.append(Graph(N=64))
                Gs.append(Graph(N=500))
                Gs.append(Graph(N=128))

        for G in Gs:
            self.assertTrue(hasattr(G, 'coords'))
            self.assertEqual(G.N, G.coords.shape[0])

            signal = np.arange(G.N) + 0.3

            G.plot(backend='pyqtgraph')
            G.plot(backend='matplotlib')
            G.plot(signal, backend='pyqtgraph')
            G.plot(signal, backend='matplotlib')
            plotting.close_all()

    def test_highlight(self):

        def test(G):
            s = np.arange(G.N)
            G.plot(s, backend='matplotlib', highlight=0)
            G.plot(s, backend='matplotlib', highlight=[0])
            G.plot(s, backend='matplotlib', highlight=[0, 1])

        # Test for 1, 2, and 3D graphs.
        G = graphs.Ring(10)
        test(G)
        G.set_coordinates('line1D')
        test(G)
        G = graphs.Torus(Nv=5)
        test(G)

    def test_indices(self):

        def test(G):
            G.plot(backend='matplotlib', indices=False)
            G.plot(backend='matplotlib', indices=True)

        # Test for 2D and 3D graphs.
        G = graphs.Ring(10)
        test(G)
        G = graphs.Torus(Nv=5)
        test(G)

    def test_signals(self):
        """Test the different kind of signals that can be plotted."""
        G = graphs.Sensor()
        G.plot()
        rng = np.random.default_rng(42)
        def test_color(param, length):
            for value in ['r', 4*(.5,), length*(2,), np.ones([1, length]),
                          rng.random(length),
                          np.ones([length, 3]), ["red"] * length,
                          rng.random([length, 4])]:
                params = {param: value}
                G.plot(**params)
            for value in [10, (0.5, 0.5), np.ones([length, 2]),
                          np.ones([2, length, 3]),
                          np.ones([length, 3]) * 1.1]:
                params = {param: value}
                self.assertRaises(ValueError, G.plot, **params)
            for value in ['r', 4*(.5)]:
                params = {param: value, 'backend': 'pyqtgraph'}
                self.assertRaises(ValueError, G.plot, **params)
        test_color('vertex_color', G.n_vertices)
        test_color('edge_color', G.n_edges)
        def test_size(param, length):
            for value in [15, length*(2,), np.ones([1, length]),
                          rng.random(length)]:
                params = {param: value}
                G.plot(**params)
            for value in [(2, 3, 4, 5), np.ones([2, length]),
                          np.ones([2, length, 3])]:
                params = {param: value}
                self.assertRaises(ValueError, G.plot, **params)
        test_size('vertex_size', G.n_vertices)
        test_size('edge_width', G.n_edges)

    def test_show_close(self):
        G = graphs.Sensor()
        G.plot()
        plotting.show(block=False)  # Don't block or the test will halt.
        plotting.close()
        plotting.close_all()

    def test_coords(self):
        G = graphs.Sensor()
        del G.coords
        self.assertRaises(AttributeError, G.plot)
        G.coords = None
        self.assertRaises(AttributeError, G.plot)
        G.coords = np.ones((G.N, 4))
        self.assertRaises(AttributeError, G.plot)
        G.coords = np.ones((G.N, 3, 1))
        self.assertRaises(AttributeError, G.plot)
        G.coords = np.ones((G.N//2, 3))
        self.assertRaises(AttributeError, G.plot)

    def test_unknown_backend(self):
        G = graphs.Sensor()
        self.assertRaises(ValueError, G.plot, backend='abc')


class TestFilters(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls._graph = graphs.Sensor(20, seed=42)
        cls._graph.compute_fourier_basis()

    def tearDown(cls):
        plotting.close_all()

    def test_all_filters(self):
        """Plot all filters."""
        for classname in dir(filters):
            if not classname[0].isupper():
                # Not a Filter class but a submodule or private stuff.
                continue
            Filter = getattr(filters, classname)
            if classname in ['Filter', 'Modulation', 'Gabor']:
                g = Filter(self._graph, filters.Heat(self._graph))
            else:
                g = Filter(self._graph)
            g.plot()
            plotting.close_all()

    def test_evaluation_points(self):
        """Change number of evaluation points."""
        def check(ax, n_lines, n_points):
            self.assertEqual(len(ax.lines), n_lines)  # n_filters + sum
            x, y = ax.lines[0].get_data()
            self.assertEqual(len(x), n_points)
            self.assertEqual(len(y), n_points)
        g = filters.Abspline(self._graph, 5)
        fig, ax = g.plot(eigenvalues=False)
        check(ax, 6, 500)
        fig, ax = g.plot(40, eigenvalues=False)
        check(ax, 6, 40)
        fig, ax = g.plot(n=20, eigenvalues=False)
        check(ax, 6, 20)

    def test_eigenvalues(self):
        """Plot with and without showing the eigenvalues."""
        graph = graphs.Sensor(20, seed=42)
        graph.estimate_lmax()
        filters.Heat(graph).plot()
        filters.Heat(graph).plot(eigenvalues=False)
        graph.compute_fourier_basis()
        filters.Heat(graph).plot()
        filters.Heat(graph).plot(eigenvalues=True)
        filters.Heat(graph).plot(eigenvalues=False)

    def test_sum_and_labels(self):
        """Plot with and without sum or labels."""
        def test(g):
            for sum in [None, True, False]:
                for labels in [None, True, False]:
                    g.plot(sum=sum, labels=labels)
        test(filters.Heat(self._graph, 10))  # one filter
        test(filters.Heat(self._graph, [10, 100]))  # multiple filters

    def test_title(self):
        """Check plot title."""
        fig, ax = filters.Wave(self._graph, 2, 1).plot()
        assert ax.get_title() == 'Wave(in=1, out=1, time=[2.00], speed=[1.00])'
        fig, ax = filters.Wave(self._graph).plot(title='test')
        assert ax.get_title() == 'test'

    def test_ax(self):
        """Axes are returned, but automatically created if not passed."""
        fig, ax = plt.subplots()
        fig2, ax2 = filters.Heat(self._graph).plot(ax=ax)
        self.assertIs(fig2, fig)
        self.assertIs(ax2, ax)

    def test_kwargs(self):
        """Additional parameters can be passed to the mpl functions."""
        g = filters.Heat(self._graph)
        g.plot(alpha=1)
        g.plot(linewidth=2)
        g.plot(linestyle='-')
        g.plot(label='myfilter')


suite = unittest.TestSuite([
    unittest.TestLoader().loadTestsFromTestCase(TestGraphs),
    unittest.TestLoader().loadTestsFromTestCase(TestFilters),
])



# -*- coding: utf-8 -*-

"""
Test suite for the utils module of the pygsp2 package.

"""

import unittest
import pandas as pd
from unidecode import unidecode

from pygsp2 import utils_examples


class TestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        pass

    @classmethod
    def tearDownClass(cls):
        pass

    def test_graphs(self):
        edgesfile='../data/santiago_metro_stations_connections.txt',
        coordsfile='../data/santiago_metro_stations_coords.geojson'
        G, pos = utils_examples.make_metro_graph(edgesfile, coordsfile)
        self.assertRaises(ValueError, utils_examples.make_metro_graph, (edgesfile, coordsfile), 'sum')

        commutes = pd.read_excel('data/2023.11 Matriz_baj_SS_MH.xlsb', header=1,
                             sheet_name='bajadas_prom_laboral')
        stations = [name.upper() for name in list(G.nodes)]
        stations = [unidecode(station) for station in stations]
        mc, s = utils_examples.metro_database_preprocessing(commutes, stations)
        fig, ax = utils_examples.plot_signal_in_graph(G, s, 'utils_examples test')


suite = unittest.TestLoader().loadTestsFromTestCase(TestCase)



# -*- coding: utf-8 -*-

import numpy as np
from scipy import sparse

from . import Graph  # prevent circular import in Python < 3.5


class Comet(Graph):
    r"""Comet graph.

    The comet is a path graph with a star of degree `k` at one end.
    Equivalently, the comet is a star made of `k` branches, where a branch of
    length `N-k` acts as the tail.
    The central vertex has degree `N-1`, the others have degree 1.

    Parameters
    ----------
    N : int
        Number of vertices.
    k : int
        Degree of central vertex.

    See Also
    --------
    Path : Comet without star
    Star : Comet without tail (path)

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> G = graphs.Comet(15, 10)
    >>> fig, axes = plt.subplots(1, 2)
    >>> _ = axes[0].spy(G.W)
    >>> _ = G.plot(ax=axes[1])

    """

    def __init__(self, N=32, k=12, **kwargs):

        if k > N-1:
            raise ValueError('The degree of the central vertex k={} must be '
                             'smaller than the number of vertices N={}.'
                             ''.format(k, N))

        self.k = k

        sources = np.concatenate((
            np.zeros(k), np.arange(k)+1,  # star
            np.arange(k, N-1), np.arange(k+1, N)  # tail (path)
        ))
        targets = np.concatenate((
            np.arange(k)+1, np.zeros(k),  # star
            np.arange(k+1, N), np.arange(k, N-1)  # tail (path)
        ))
        n_edges = N - 1
        weights = np.ones(2*n_edges)
        W = sparse.csr_matrix((weights, (sources, targets)), shape=(N, N))

        indices = np.arange(k) + 1
        coords = np.zeros((N, 2))
        coords[1:k+1, 0] = np.cos(indices*2*np.pi/k)
        coords[1:k+1, 1] = np.sin(indices*2*np.pi/k)
        coords[k+1:, 0] = np.arange(1, N-k) + 1

        super(Comet, self).__init__(W, coords=coords, **kwargs)

    def _get_extra_repr(self):
        return dict(k=self.k)



# -*- coding: utf-8 -*-

from . import Comet  # prevent circular import in Python < 3.5


class Star(Comet):
    r"""Star graph.

    A star with a central vertex and `N-1` branches.
    The central vertex has degree `N-1`, the others have degree 1.

    Parameters
    ----------
    N : int
        Number of vertices.

    See Also
    --------
    Comet : Generalization with a longer branch as a tail

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> graph = graphs.Star(15)
    >>> graph
    Star(n_vertices=15, n_edges=14)
    >>> fig, axes = plt.subplots(1, 2)
    >>> _ = axes[0].spy(graph.W)
    >>> _ = graph.plot(ax=axes[1])

    """

    def __init__(self, N=10, **kwargs):
        plotting = dict(limits=[-1.1, 1.1, -1.1, 1.1])
        plotting.update(kwargs.get('plotting', {}))
        super(Star, self).__init__(N, N-1, plotting=plotting, **kwargs)

    def _get_extra_repr(self):
        return dict()  # Suppress Comet repr.



# -*- coding: utf-8 -*-

r"""
The :mod:`pygsp2.optimization` module provides tools to solve convex
optimization problems on graphs.
"""

from pygsp2 import utils


logger = utils.build_logger(__name__)


def _import_pyunlocbox():
    try:
        from pyunlocbox import functions, solvers
    except Exception as e:
        raise ImportError('Cannot import pyunlocbox, which is needed to solve '
                          'this optimization problem. Try to install it with '
                          'pip (or conda) install pyunlocbox. '
                          'Original exception: {}'.format(e))
    return functions, solvers


def prox_tv(x, gamma, G, A=None, At=None, nu=1, tol=10e-4, maxit=200, use_matrix=True):
    r"""
    Total Variation proximal operator for graphs.

    This function computes the TV proximal operator for graphs. The TV norm
    is the one norm of the gradient. The gradient is defined in the
    function :meth:`pygsp2.graphs.Graph.grad`.
    This function requires the PyUNLocBoX to be executed.

    This function solves:

    :math:`sol = \min_{z} \frac{1}{2} \|x - z\|_2^2 + \gamma  \|x\|_{TV}`

    Parameters
    ----------
    x: int
        Input signal
    gamma: ndarray
        Regularization parameter
    G: graph object
        Graphs structure
    A: lambda function
        Forward operator, this parameter allows to solve the following problem:
        :math:`sol = \min_{z} \frac{1}{2} \|x - z\|_2^2 + \gamma  \| A x\|_{TV}`
        (default = Id)
    At: lambda function
        Adjoint operator. (default = Id)
    nu: float
        Bound on the norm of the operator (default = 1)
    tol: float
        Stops criterion for the loop. The algorithm will stop if :
        :math:`\frac{n(t) - n(t - 1)} {n(t)} < tol`
        where :math:`n(t) = f(x) + 0.5 \|x-y\|_2^2` is the objective function at iteration :math:`t`
        (default = :math:`10e-4`)
    maxit: int
        Maximum iteration. (default = 200)
    use_matrix: bool
        If a matrix should be used. (default = True)

    Returns
    -------
    sol: solution

    Examples
    --------

    """
    if A is None:
        def A(x):
            return x
    if At is None:
        def At(x):
            return x

    tight = 0
    l1_nu = 2 * G.lmax * nu

    if use_matrix:
        def l1_a(x):
            return G.Diff * A(x)

        def l1_at(x):
            return G.Diff * At(D.T * x)
    else:
        def l1_a(x):
            return G.grad(A(x))

        def l1_at(x):
            return G.div(x)

    functions, _ = _import_pyunlocbox()
    functions.norm_l1(x, gamma, A=l1_a, At=l1_at, tight=tight, maxit=maxit, verbose=verbose, tol=tol)



# -*- coding: utf-8 -*-

import numpy as np

from pygsp2 import utils
from pygsp2.graphs import NNGraph  # prevent circular import in Python < 3.5


class TwoMoons(NNGraph):
    r"""Two Moons (NN-graph).

    Parameters
    ----------
    moontype : 'standard' or 'synthesized'
        You have the freedom to chose if you want to create a standard
        two_moons graph or a synthesized one (default is 'standard').
        'standard' : Create a two_moons graph from a based graph.
        'synthesized' : Create a synthesized two_moon
    sigmag : float
        Variance of the distance kernel (default = 0.05)
    dim : int
        The dimensionality of the points (default = 2).
        Only valid for moontype == 'standard'.
    N : int
        Number of vertices (default = 2000)
        Only valid for moontype == 'synthesized'.
    sigmad : float
        Variance of the data (do not set it too high or you won't see anything)
        (default = 0.05)
        Only valid for moontype == 'synthesized'.
    distance : float
        Distance between the two moons (default = 0.5)
        Only valid for moontype == 'synthesized'.
    seed : int
        Seed for the random number generator (for reproducible graphs).

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> G = graphs.TwoMoons()
    >>> fig, axes = plt.subplots(1, 2)
    >>> _ = axes[0].spy(G.W, markersize=0.5)
    >>> _ = G.plot(edges=True, ax=axes[1])

    """

    def _create_arc_moon(self, N, sigmad, distance, number, seed):
        rng = np.random.default_rng(seed)
        phi = rng.uniform(size=(N, 1)) * np.pi
        r = 1
        rb = sigmad * rng.normal(size=(N, 1))
        ab = rng.uniform(size=(N, 1)) * 2 * np.pi
        b = rb * np.exp(1j * ab)
        bx = np.real(b)
        by = np.imag(b)

        if number == 1:
            moonx = np.cos(phi) * r + bx + 0.5
            moony = -np.sin(phi) * r + by - (distance - 1)/2.
        elif number == 2:
            moonx = np.cos(phi) * r + bx - 0.5
            moony = np.sin(phi) * r + by + (distance - 1)/2.

        return np.concatenate((moonx, moony), axis=1)

    def __init__(self, moontype='standard', dim=2, sigmag=0.05,
                 N=400, sigmad=0.07, distance=0.5, seed=None, **kwargs):

        self.moontype = moontype
        self.dim = dim
        self.sigmag = sigmag
        self.sigmad = sigmad
        self.distance = distance
        self.seed = seed

        if moontype == 'standard':
            N1, N2 = 1000, 1000
            data = utils.loadmat('pointclouds/two_moons')
            Xin = data['features'][:dim].T

        elif moontype == 'synthesized':
            N1 = N // 2
            N2 = N - N1

            coords1 = self._create_arc_moon(N1, sigmad, distance, 1, seed)
            coords2 = self._create_arc_moon(N2, sigmad, distance, 2, seed)

            Xin = np.concatenate((coords1, coords2))

        else:
            raise ValueError('Unknown moontype {}'.format(moontype))

        self.labels = np.concatenate((np.zeros(N1), np.ones(N2)))

        plotting = {
            'vertex_size': 30,
        }

        super(TwoMoons, self).__init__(Xin=Xin, sigma=sigmag, k=5,
                                       center=False, rescale=False,
                                       plotting=plotting, **kwargs)

    def _get_extra_repr(self):
        attrs = {'moontype': self.moontype,
                 'dim': self.dim,
                 'sigmag': '{:.2f}'.format(self.sigmag),
                 'sigmad': '{:.2f}'.format(self.sigmad),
                 'distance': '{:.2f}'.format(self.distance),
                 'seed': self.seed}
        attrs.update(super(TwoMoons, self)._get_extra_repr())
        return attrs



# -*- coding: utf-8 -*-

import pygsp2

extensions = [
    'sphinx.ext.viewcode',
    'sphinx.ext.autosummary',
    'sphinx.ext.mathjax',
    'sphinx.ext.inheritance_diagram',
]

extensions.append('sphinx.ext.autodoc')
autodoc_default_options = {
    'members': True,
    'undoc-members': True,
    'member-order': 'groupwise',  # alphabetical, groupwise, bysource
}

extensions.append('sphinx.ext.intersphinx')
intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
    'numpy': ('https://numpy.org/doc/stable', None),
    'scipy': ('https://docs.scipy.org/doc/scipy/reference', None),
    'matplotlib': ('https://matplotlib.org/stable', None),
    'pyunlocbox': ('https://pyunlocbox.readthedocs.io/en/stable', None),
    'networkx': ('https://networkx.org/documentation/stable', None),
    'graph_tool': ('https://graph-tool.skewed.de/static/doc', None),
}

extensions.append('numpydoc')
numpydoc_show_class_members = False
numpydoc_use_plots = True  # Add the plot directive whenever mpl is imported.

extensions.append('matplotlib.sphinxext.plot_directive')
plot_include_source = True
plot_html_show_source_link = False
plot_html_show_formats = False
plot_working_directory = '.'
plot_rcparams = {
    'figure.figsize': (10, 4)
}
plot_pre_code = """
import numpy as np
from pygsp2 import graphs, filters, utils, plotting
"""

extensions.append('sphinx_gallery.gen_gallery')
sphinx_gallery_conf = {
    'examples_dirs': '../examples',
    'gallery_dirs': 'examples',
    'filename_pattern': '/',
    'reference_url': {'pygsp2': None},
    'backreferences_dir': 'backrefs',
    'doc_module': 'pygsp2',
    'show_memory': True,
}

extensions.append('sphinx_copybutton')
copybutton_prompt_text = ">>> "

extensions.append('sphinxcontrib.bibtex')
bibtex_bibfiles = ['references.bib']

exclude_patterns = ['_build']
source_suffix = '.rst'
master_doc = 'index'

project = 'pygsp2'
version = pygsp2.__version__
release = pygsp2.__version__
copyright = 'EPFL LTS2'

pygments_style = 'sphinx'
html_theme = 'sphinx_rtd_theme'
html_theme_options = {
    'navigation_depth': 2,
}
latex_elements = {
    'papersize': 'a4paper',
    'pointsize': '10pt',
}
latex_documents = [
    ('index', 'pygsp2.tex', 'PyGSP2 documentation',
     'EPFL LTS2', 'manual'),
]



# -*- coding: utf-8 -*-

"""
Test suite for the docstrings of the pygsp2 package.

"""

import os
import unittest
import doctest


def gen_recursive_file(root, ext):
    for root, _, filenames in os.walk(root):
        for name in filenames:
            if name.lower().endswith(ext):
                yield os.path.join(root, name)


def test_docstrings(root, ext, setup=None):
    files = list(gen_recursive_file(root, ext))
    return doctest.DocFileSuite(*files, setUp=setup, tearDown=teardown,
                                module_relative=False)


def setup(doctest):
    import numpy
    import pygsp2
    doctest.globs = {
        'graphs': pygsp2.graphs,
        'filters': pygsp2.filters,
        'utils': pygsp2.utils,
        'np': numpy,
    }


def teardown(doctest):
    """Close matplotlib figures to avoid warning and save memory."""
    import pygsp2
    pygsp2.plotting.close_all()


# Docstrings from API reference.
suite_reference = test_docstrings('pygsp2', '.py', setup)

# Docstrings from tutorials. No setup to not forget imports.
suite_tutorials = test_docstrings('.', '.rst')

suite = unittest.TestSuite([suite_reference, suite_tutorials])



# -*- coding: utf-8 -*-

from __future__ import division

import numpy as np

from . import Filter  # prevent circular import in Python < 3.5


class Regular(Filter):
    r"""Design 2 filters with the regular construction (tight frame).

    This function creates a Parseval filter bank of 2 filters.
    The low-pass filter is defined by a function :math:`f_l(x)`
    between :math:`0` and :math:`2`. For :math:`d = 0`.

    .. math:: f_{l}= \sin\left( \frac{\pi}{4} x \right)

    For :math:`d = 1`

    .. math:: f_{l}= \sin\left( \frac{\pi}{4} \left( 1+ \sin\left(\frac{\pi}{2}(x-1)\right) \right) \right)

    For :math:`d = 2`

    .. math:: f_{l}= \sin\left( \frac{\pi}{4} \left( 1+ \sin\left(\frac{\pi}{2} \sin\left(\frac{\pi}{2}(x-1)\right)\right) \right) \right)

    And so forth for other degrees :math:`d`.

    The high pass filter is adapted to obtain a tight frame.

    Parameters
    ----------
    G : graph
    degree : float
        Degree (default = 3). See above equations.

    Examples
    --------

    Filter bank's representation in Fourier and time (ring graph) domains.

    >>> import matplotlib.pyplot as plt
    >>> G = graphs.Ring(N=20)
    >>> G.estimate_lmax()
    >>> G.set_coordinates('line1D')
    >>> g = filters.Regular(G)
    >>> s = g.localize(G.N // 2)
    >>> fig, axes = plt.subplots(1, 2)
    >>> _ = g.plot(ax=axes[0])
    >>> _ = G.plot(s, ax=axes[1])

    """

    def __init__(self, G, degree=3):

        self.degree = degree

        def kernel(x, degree):
            if degree == 0:
                return np.sin(np.pi / 4 * x)
            else:
                output = np.sin(np.pi * (x - 1) / 2)
                for _ in range(2, degree):
                    output = np.sin(np.pi * output / 2)
                return np.sin(np.pi / 4 * (1 + output))

        regular = Filter(G, lambda x: kernel(x*2/G.lmax, degree))
        complement = regular.complement(frame_bound=1)
        kernels = regular._kernels + complement._kernels

        super(Regular, self).__init__(G, kernels)



r"""
Localization of Fourier modes
=============================

The Fourier modes (the eigenvectors of the graph Laplacian) can be localized in
the spacial domain. As a consequence, graph signals can be localized in both
space and frequency (which is impossible for Euclidean domains or manifolds, by
the Heisenberg's uncertainty principle).

This example demonstrates that the more isolated a node is, the more a Fourier
mode will be localized on it.

The mutual coherence between the basis of Kronecker deltas and the basis formed
by the eigenvectors of the Laplacian, :attr:`pygsp2.graphs.Graph.coherence`, is
a measure of the localization of the Fourier modes. The larger the value, the
more localized the eigenvectors can be.

See `Global and Local Uncertainty Principles for Signals on Graphs
<https://arxiv.org/abs/1603.03030>`_ for details.
"""

import numpy as np
import matplotlib as mpl
from matplotlib import pyplot as plt
import pygsp2 as pg

fig, axes = plt.subplots(2, 2, figsize=(8, 8))

for w, ax in zip([10, 1, 0.1, 0.01], axes.flatten()):

    adjacency = [
        [0, w, 0, 0],
        [w, 0, 1, 0],
        [0, 1, 0, 1],
        [0, 0, 1, 0],
    ]
    graph = pg.graphs.Graph(adjacency)
    graph.compute_fourier_basis()

    # Plot eigenvectors.
    ax.plot(graph.U)
    ax.set_ylim(-1, 1)
    ax.set_yticks([-1, 0, 1])
    ax.legend([f'$u_{i}(v)$, $\lambda_{i}={graph.e[i]:.1f}$' for i in
              range(graph.n_vertices)], loc='upper right')

    ax.text(0, -0.9, f'coherence = {graph.coherence:.2f}'
            f'$\in [{1/np.sqrt(graph.n_vertices)}, 1]$')

    # Plot vertices.
    ax.set_xticks(range(graph.n_vertices))
    ax.set_xticklabels([f'$v_{i}$' for i in range(graph.n_vertices)])

    # Plot graph.
    x, y = np.arange(0, graph.n_vertices), -1.20*np.ones(graph.n_vertices)
    line = mpl.lines.Line2D(x, y, lw=3, color='k', marker='.', markersize=20)
    line.set_clip_on(False)
    ax.add_line(line)

    # Plot edge weights.
    for i in range(graph.n_vertices - 1):
        j = i+1
        ax.text(i+0.5, -1.15, f'$w_{{{i}{j}}} = {adjacency[i][j]}$',
                horizontalalignment='center')

fig.tight_layout()
plt.show()


# -*- coding: utf-8 -*-

import numpy as np

from pygsp2 import utils
from . import Filter  # prevent circular import in Python < 3.5


class SimpleTight(Filter):
    r"""Design a simple tight frame filter bank (tight frame).

    These filters have been designed to be a simple tight frame wavelet filter
    bank. The kernel is similar to Meyer, but simpler. The function is
    essentially :math:`\sin^2(x)` in ascending part and :math:`\cos^2` in
    descending part.

    Parameters
    ----------
    G : graph
    Nf : int
        Number of filters to cover the interval [0, lmax].
    scales : array_like
        Scales to be used. Defaults to log scale.

    Examples
    --------

    Filter bank's representation in Fourier and time (ring graph) domains.

    >>> import matplotlib.pyplot as plt
    >>> G = graphs.Ring(N=20)
    >>> G.estimate_lmax()
    >>> G.set_coordinates('line1D')
    >>> g = filters.SimpleTight(G)
    >>> s = g.localize(G.N // 2)
    >>> fig, axes = plt.subplots(1, 2)
    >>> _ = g.plot(ax=axes[0])
    >>> _ = G.plot(s, ax=axes[1])

    """

    def __init__(self, G, Nf=6, scales=None):

        def kernel(x, kerneltype):
            r"""
            Evaluates 'simple' tight-frame kernel.

            * simple tf wavelet kernel: supported on [1/4, 1]
            * simple tf scaling function kernel: supported on [0, 1/2]

            Parameters
            ----------
            x : ndarray
                Array of independent variable values
            kerneltype : str
                Can be either 'sf' or 'wavelet'

            Returns
            -------
            r : ndarray

            """

            l1 = 0.25
            l2 = 0.5
            l3 = 1.

            def h(x):
                return np.sin(np.pi*x/2.)**2

            r1ind = (x < l1)
            r2ind = (x >= l1) * (x < l2)
            r3ind = (x >= l2) * (x < l3)

            r = np.zeros(x.shape)
            if kerneltype == 'sf':
                r[r1ind] = 1.
                r[r2ind] = np.sqrt(1 - h(4*x[r2ind] - 1)**2)
            elif kerneltype == 'wavelet':
                r[r2ind] = h(4*(x[r2ind] - 1/4.))
                r[r3ind] = np.sqrt(1 - h(2*x[r3ind] - 1)**2)
            else:
                raise TypeError('Unknown kernel type', kerneltype)

            return r

        if not scales:
            scales = (1./(2.*G.lmax) * np.power(2, np.arange(Nf-2, -1, -1)))
        self.scales = scales

        if len(scales) != Nf - 1:
            raise ValueError('len(scales) should be Nf-1.')

        kernels = [lambda x: kernel(scales[0] * x, 'sf')]

        for i in range(Nf - 1):
            kernels.append(lambda x, i=i: kernel(scales[i] * x, 'wavelet'))

        super(SimpleTight, self).__init__(G, kernels)



r"""
Filtering a signal
==================

A graph signal is filtered by transforming it to the spectral domain (via the
Fourier transform), performing a point-wise multiplication (motivated by the
convolution theorem), and transforming it back to the vertex domain (via the
inverse graph Fourier transform).

.. note::

    In practice, filtering is implemented in the vertex domain to avoid the
    computationally expensive graph Fourier transform. To do so, filters are
    implemented as polynomials of the eigenvalues / Laplacian. Hence, filtering
    a signal reduces to its multiplications with sparse matrices (the graph
    Laplacian).

"""

import numpy as np
from matplotlib import pyplot as plt
import pygsp2 as pg

G = pg.graphs.Sensor(seed=42)
G.compute_fourier_basis()

#g = pg.filters.Rectangular(G, band_max=0.2)
g = pg.filters.Expwin(G, band_max=0.5)

fig, axes = plt.subplots(1, 3, figsize=(12, 4))
fig.subplots_adjust(hspace=0.5)

x = np.random.default_rng(1).normal(size=G.N)
#x = np.random.default_rng(42).uniform(-1, 1, size=G.N)
x = 3 * x / np.linalg.norm(x)
y = g.filter(x)
x_hat = G.gft(x).squeeze()
y_hat = G.gft(y).squeeze()

limits = [x.min(), x.max()]

G.plot(x, limits=limits, ax=axes[0], title='input signal $x$ in the vertex domain')
axes[0].text(0, -0.1, '$x^T L x = {:.2f}$'.format(G.dirichlet_energy(x)))
axes[0].set_axis_off()

g.plot(ax=axes[1], alpha=1)
line_filt = axes[1].lines[-2]
line_in, = axes[1].plot(G.e, np.abs(x_hat), '.-')
line_out, = axes[1].plot(G.e, np.abs(y_hat), '.-')
#axes[1].set_xticks(range(0, 16, 4))
axes[1].set_xlabel(r'graph frequency $\lambda$')
axes[1].set_ylabel(r'frequency content $\hat{x}(\lambda)$')
axes[1].set_title(r'signals in the spectral domain')
axes[1].legend(['input signal $\hat{x}$'])
labels = [
    r'input signal $\hat{x}$',
    'kernel $g$',
    r'filtered signal $\hat{y}$',
]
axes[1].legend([line_in, line_filt, line_out], labels, loc='upper right')

G.plot(y, limits=limits, ax=axes[2], title='filtered signal $y$ in the vertex domain')
axes[2].text(0, -0.1, '$y^T L y = {:.2f}$'.format(G.dirichlet_energy(y)))
axes[2].set_axis_off()

fig.tight_layout()
plt.show()


# -*- coding: utf-8 -*-

import numpy as np
from scipy import sparse

from pygsp2 import utils
from . import Graph  # prevent circular import in Python < 3.5


class RandomRing(Graph):
    r"""Ring graph with randomly sampled vertices.

    Parameters
    ----------
    N : int
        Number of vertices.
    angles : array_like, optional
        The angular coordinate, in :math:`[0, 2\pi]`, of the vertices.
    seed : int
        Seed for the random number generator (for reproducible graphs).

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> G = graphs.RandomRing(N=10, seed=42)
    >>> fig, axes = plt.subplots(1, 2)
    >>> _ = axes[0].spy(G.W)
    >>> _ = G.plot(ax=axes[1])
    >>> _ = axes[1].set_xlim(-1.1, 1.1)
    >>> _ = axes[1].set_ylim(-1.1, 1.1)

    """

    def __init__(self, N=64, angles=None, seed=None, **kwargs):

        self.seed = seed

        if angles is None:
            rng = np.random.default_rng(seed)
            angles = np.sort(rng.uniform(0, 2*np.pi, size=N), axis=0)
        else:
            angles = np.asanyarray(angles)
            angles.sort()  # Need to be sorted to take the difference.
            N = len(angles)
            if np.any(angles < 0) or np.any(angles >= 2*np.pi):
                raise ValueError('Angles should be in [0, 2 pi]')
        self.angles = angles

        if N < 3:
            # Asymmetric graph needed for 2 as 2 distances connect them.
            raise ValueError('There should be at least 3 vertices.')

        rows = range(0, N-1)
        cols = range(1, N)
        weights = np.diff(angles)

        # Close the loop.
        rows = np.concatenate((rows, [0]))
        cols = np.concatenate((cols, [N-1]))
        weights = np.concatenate((weights, [2*np.pi + angles[0] - angles[-1]]))

        W = sparse.coo_matrix((weights, (rows, cols)), shape=(N, N))
        W = utils.symmetrize(W, method='triu')

        # Width as the expected angle. All angles are equal to that value when
        # the ring is uniformly sampled.
        width = 2 * np.pi / N
        assert (W.data.mean() - width) < 1e-10
        # TODO: why this kernel ? It empirically produces eigenvectors closer
        # to the sines and cosines.
        W.data = width / W.data

        coords = np.stack([np.cos(angles), np.sin(angles)], axis=1)
        plotting = {'limits': np.array([-1, 1, -1, 1])}

        # TODO: save angle and 2D position as graph signals
        super(RandomRing, self).__init__(W, coords=coords, plotting=plotting,
                                         **kwargs)

    def _get_extra_repr(self):
        return dict(seed=self.seed)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup


setup(
    name='pygsp2',
    version='v2.0.10',
    description='Graph Signal Processing in Python',
    long_description=open('README.rst').read(),
    long_description_content_type='text/x-rst',
    author='EPFL LTS2',
    url='https://github.com/gsp-eeg/pygsp2',
    project_urls={
        'Documentation': 'https://pygsp2.readthedocs.io',
        'Download': 'https://pypi.org/project/pygsp2',
        'Source Code': 'https://github.com/gsp-eeg/pygsp2',
        'Bug Tracker': 'https://github.com/gsp-eeg/pygsp2/issues',
        'Try It Online': 'https://mybinder.org/v2/gh/epfl-lts2/pygsp2/master?urlpath=lab/tree/examples/playground.ipynb',
    },
    packages=[
        'pygsp2',
        'pygsp2.graphs',
        'pygsp2.graphs.nngraphs',
        'pygsp2.filters',
        'pygsp2.tests',
    ],
    package_data={'pygsp2': ['data/pointclouds/*.mat']},
    test_suite='pygsp2.tests.suite',
    python_requires='>=3.7',
    install_requires=[
        'numpy',
        'scipy',
        'pandas',
        'networkx',
        'geopy',
        'matplotlib',
        'unidecode',
        'utm',
        'pyxlsb',
        'charset-normalizer'
    ],
    extras_require={
        # Optional dependencies for development. Some bring additional
        # functionalities, others are for testing, documentation, or packaging.
        'dev': [
            # Import and export.
            'networkx',
            #'json',
            'utm',
            'geopy',
            'pyxlsb',
            'unidecode',
            # 'graph-tool', cannot be installed by pip
            # Construct patch graphs from images.
            'scikit-image',
            # Approximate nearest neighbors for kNN graphs.
            'pyflann3',
            # Convex optimization on graph.
            'pyunlocbox',
            # Plot graphs, signals, and filters.
            'matplotlib',
            # Interactive graph visualization.
            'pyqtgraph',
            'PyOpenGL',
            'PyQt5',
            # Run the tests.
            'flake8',
            'coverage',
            'coveralls',
            # Build the documentation.
            'sphinx',
            'numpydoc',
            'sphinxcontrib-bibtex',
            'sphinx-gallery',
            'memory_profiler',
            'sphinx-rtd-theme',
            'sphinx-copybutton',
            # Build and upload packages.
            'wheel',
            'twine',
        ],
    },
    license="BSD",
    keywords='graph signal processing',
    platforms='any',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Topic :: Scientific/Engineering',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
)



# -*- coding: utf-8 -*-

r"""
The :mod:`pygsp2.filters` module implements methods used for filtering and
defines commonly used filters that can be applied to :mod:`pygsp2.graphs`. A
filter is associated to a graph and is defined with one or several functions.
We define by filter bank a list of filters, usually centered around different
frequencies, applied to a single graph.

Interface
---------

The :class:`Filter` base class implements a common interface to all filters:

.. autosummary::

    Filter.evaluate
    Filter.filter
    Filter.analyze
    Filter.synthesize
    Filter.complement
    Filter.inverse
    Filter.compute_frame
    Filter.estimate_frame_bounds
    Filter.plot
    Filter.localize

Filters
-------

Then, derived classes implement various common graph filters.

**Filters that solve differential equations**

The following filters solve partial differential equations (PDEs) on graphs,
which model processes such as heat diffusion or wave propagation.

.. autosummary::

    Heat
    Wave

**Low-pass filters**

.. autosummary::

    Heat

**Band-pass filters**

These filters can be configured to be low-pass, high-pass, or band-pass.

.. autosummary::

    Expwin
    Rectangular

**Filter banks of two filters: a low-pass and a high-pass**

.. autosummary::

    Regular
    Held
    Simoncelli
    Papadakis

**Filter banks composed of dilated or translated filters**

.. autosummary::

    Abspline
    HalfCosine
    Itersine
    MexicanHat
    Meyer
    SimpleTight

**Filter banks for vertex-frequency analyzes**

Those filter banks are composed of shifted versions of a mother filter, one per
graph frequency (Laplacian eigenvalue). They can analyze frequency content
locally, as a windowed graph Fourier transform.

.. autosummary::

    Gabor
    Modulation

Approximations
--------------

Moreover, two approximation methods are provided for fast filtering. The
computational complexity of filtering with those approximations is linear with
the number of edges. The complexity of the exact solution, which is to use the
Fourier basis, is quadratic with the number of nodes (without taking into
account the cost of the necessary eigendecomposition of the graph Laplacian).

**Chebyshev polynomials**

.. autosummary::

    compute_cheby_coeff
    compute_jackson_cheby_coeff
    cheby_op
    cheby_rect

**Lanczos algorithm**

.. autosummary::

    lanczos
    lanczos_op

"""

from .filter import Filter  # noqa: F401
from .abspline import Abspline  # noqa: F401
from .expwin import Expwin  # noqa: F401
from .gabor import Gabor  # noqa: F401
from .halfcosine import HalfCosine  # noqa: F401
from .heat import Heat  # noqa: F401
from .held import Held  # noqa: F401
from .itersine import Itersine  # noqa: F401
from .mexicanhat import MexicanHat  # noqa: F401
from .meyer import Meyer  # noqa: F401
from .modulation import Modulation  # noqa: F401
from .papadakis import Papadakis  # noqa: F401
from .rectangular import Rectangular  # noqa: F401
from .regular import Regular  # noqa: F401
from .simoncelli import Simoncelli  # noqa: F401
from .simpletight import SimpleTight  # noqa: F401
from .wave import Wave  # noqa: F401

from .approximations import compute_cheby_coeff  # noqa: F401
from .approximations import compute_jackson_cheby_coeff  # noqa: F401
from .approximations import cheby_op  # noqa: F401
from .approximations import cheby_rect  # noqa: F401
from .approximations import lanczos  # noqa: F401
from .approximations import lanczos_op  # noqa: F401



# -*- coding: utf-8 -*-

import numpy as np
from scipy import sparse

from pygsp2 import utils
from . import Graph  # prevent circular import in Python < 3.5


class Grid2d(Graph):
    r"""2-dimensional grid graph.

    On the 2D grid, the graph Fourier transform (GFT) is the Kronecker product
    between the GFT of two :class:`~pygsp2.graphs.Path` graphs.

    Parameters
    ----------
    N1 : int
        Number of vertices along the first dimension.
    N2 : int
        Number of vertices along the second dimension. Default is ``N1``.
    diagonal : float
        Value of the diagnal edges. Default is ``0.0``

    See Also
    --------
    Path : 1D line with even boundary conditions
    Torus : Kronecker product of two ring graphs
    Grid2dImgPatches

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> G = graphs.Grid2d(N1=5, N2=4)
    >>> fig, axes = plt.subplots(1, 2)
    >>> _ = axes[0].spy(G.W)
    >>> _ = G.plot(ax=axes[1])

    """

    def __init__(self, N1=16, N2=None, diagonal=0.0, **kwargs):

        if N2 is None:
            N2 = N1

        self.N1 = N1
        self.N2 = N2

        N = N1 * N2

        # Filling up the weight matrix this way is faster than
        # looping through all the grid points:
        diag_1 = np.ones(N - 1)
        diag_1[(N2 - 1)::N2] = 0
        diag_2 = np.ones(N - N2)

        W = sparse.diags(diagonals=[diag_1, diag_2],
                         offsets=[-1, -N2],
                         shape=(N, N),
                         format='csr',
                         dtype='float')

        if min(N1, N2) > 1 and diagonal != 0.0:
            # Connecting node with they diagonal neighbours
            diag_3 = np.full(N - N2 - 1, diagonal)
            diag_4 = np.full(N - N2 + 1, diagonal)
            diag_3[N2 - 1::N2] = 0
            diag_4[0::N2] = 0
            D = sparse.diags(diagonals=[diag_3, diag_4],
                             offsets=[-N2 - 1, -N2 + 1],
                             shape=(N, N),
                             format='csr',
                             dtype='float')
            W += D

        W = utils.symmetrize(W, method='tril')

        x = np.kron(np.ones((N1, 1)), (np.arange(N2)/float(N2)).reshape(N2, 1))
        y = np.kron(np.ones((N2, 1)), np.arange(N1)/float(N1)).reshape(N, 1)
        y = np.sort(y, axis=0)[::-1]
        coords = np.concatenate((x, y), axis=1)

        plotting = {"limits": np.array([-1. / N2, 1 + 1. / N2,
                                        1. / N1, 1 + 1. / N1])}

        super(Grid2d, self).__init__(W, coords=coords,
                                     plotting=plotting, **kwargs)

    def _get_extra_repr(self):
        return dict(N1=self.N1, N2=self.N2)



r"""
Heat diffusion
==============

Solve the heat equation by filtering the initial conditions with the heat
kernel :class:`pygsp2.filters.Heat`.
"""

from os import path

import numpy as np
from matplotlib import pyplot as plt
import pygsp2 as pg

n_side = 13
G = pg.graphs.Grid2d(n_side)
G.compute_fourier_basis()

sources = [
    (n_side//4 * n_side) + (n_side//4),
    (n_side*3//4 * n_side) + (n_side*3//4),
]
x = np.zeros(G.n_vertices)
x[sources] = 5

times = [0, 5, 10, 20]

fig, axes = plt.subplots(2, len(times), figsize=(12, 5))
for i, t in enumerate(times):
    g = pg.filters.Heat(G, scale=t)
    title = r'$\hat{{f}}({0}) = g_{{1,{0}}} \odot \hat{{f}}(0)$'.format(t)
    g.plot(alpha=1, ax=axes[0, i], title=title)
    axes[0, i].set_xlabel(r'$\lambda$')
#    axes[0, i].set_ylabel(r'$g(\lambda)$')
    if i > 0:
        axes[0, i].set_ylabel('')
    y = g.filter(x)
    line, = axes[0, i].plot(G.e, G.gft(y))
    labels = [r'$\hat{{f}}({})$'.format(t), r'$g_{{1,{}}}$'.format(t)]
    axes[0, i].legend([line, axes[0, i].lines[-3]], labels, loc='lower right')
    G.plot(y, edges=False, highlight=sources, ax=axes[1, i], title=r'$f({})$'.format(t))
    axes[1, i].set_aspect('equal', 'box')
    axes[1, i].set_axis_off()

fig.tight_layout()
plt.show()


r"""
Graph learning
==============
Use Kalofolias2016 algorithm to learn the graph from a signal.
The algorithm uses the pairwise distances between each node signal.
In this case, we compute the euclidean distace between the node's coordenates
but the euclidean distance between each node signal would work in the same style.
"""
# %% Import Libraries
import numpy as np
from scipy import spatial
import matplotlib.pyplot as plt
from pygsp2 import graphs, graph_learning

# %% Create a ring graph
G = graphs.ring.Ring(10)

# We will learn the graph from the coords
# but signals from the graph also work the same
signal = G.coords

# %% Compute distance

kdt = spatial.KDTree(signal)

D, NN = kdt.query(signal[:, None], k=len(signal))

# Allocate distance array
Z = np.zeros((G.N, G.N))
for i, n in enumerate(NN):
    Z[i, n] = D[i]

# Learn graph
A = 0.8
B = 0.1

W = graph_learning.graph_log_degree(Z, A, B)
W[W < 1e-5] = 0

plt.figure(figsize=(10, 4))
plt.subplot(121)
plt.imshow(G.W.toarray(), vmin=0, vmax=1)
plt.colorbar()
plt.title('Original Graph')

plt.subplot(122)
plt.imshow(W, vmin=0, vmax=1)
plt.colorbar()
plt.title('Learned Graph')

plt.show()



# -*- coding: utf-8 -*-

from __future__ import division

import numpy as np

from . import Filter  # prevent circular import in Python < 3.5


class Expwin(Filter):
    r"""Design an exponential window filter.

    The window has the shape of a (half) triangle with the corners smoothed by
    an exponential function.

    Parameters
    ----------
    G : graph
    band_min : float
        Minimum relative band. The filter evaluates at 0.5 at this frequency.
        Zero corresponds to the smallest eigenvalue (which is itself equal to
        zero), one corresponds to the largest eigenvalue.
        If None, the filter is high-pass.
    band_max : float
        Maximum relative band. The filter evaluates at 0.5 at this frequency.
        If None, the filter is low-pass.
    slope : float
        The slope at cut-off.

    Examples
    --------

    Filter bank's representation in Fourier and time (ring graph) domains.

    >>> import matplotlib.pyplot as plt
    >>> G = graphs.Ring(N=20)
    >>> G.estimate_lmax()
    >>> G.set_coordinates('line1D')
    >>> g = filters.Expwin(G, band_min=0.1, band_max=0.7, slope=5)
    >>> s = g.localize(G.N // 2)
    >>> fig, axes = plt.subplots(1, 2)
    >>> _ = g.plot(ax=axes[0])
    >>> _ = G.plot(s, ax=axes[1])

    """

    def __init__(self, G, band_min=None, band_max=0.2, slope=1):

        self.band_min = band_min
        self.band_max = band_max
        self.slope = slope

        def exp(x):
            """Exponential function with canary to avoid division by zero and
            overflow."""
            y = np.where(x <= 0, -1, x)
            y = np.exp(-slope / y)
            return np.where(x <= 0, 0, y)

        def h(x):
            y = exp(x)
            z = exp(1 - x)
            return y / (y + z)

        def kernel_lowpass(x):
            return h(0.5 - x/G.lmax + band_max)
        def kernel_highpass(x):
            return h(0.5 + x/G.lmax - band_min)

        if (band_min is None) and (band_max is None):
            kernel = lambda x: np.ones_like(x)
        elif band_min is None:
            kernel = kernel_lowpass
        elif band_max is None:
            kernel = kernel_highpass
        else:
            kernel = lambda x: kernel_lowpass(x) * kernel_highpass(x)

        super(Expwin, self).__init__(G, kernel)

    def _get_extra_repr(self):
        attrs = dict()
        if self.band_min is not None:
            attrs.update(band_min='{:.2f}'.format(self.band_min))
        if self.band_max is not None:
            attrs.update(band_max='{:.2f}'.format(self.band_max))
        attrs.update(slope='{:.0f}'.format(self.slope))
        return attrs



# -*- coding: utf-8 -*-

# prevent circular import in Python < 3.5
from .stochasticblockmodel import StochasticBlockModel


class ErdosRenyi(StochasticBlockModel):
    r"""Erdos Renyi graph.

    The Erdos Renyi graph is constructed by randomly connecting nodes. Each
    edge is included in the graph with probability p, independently from any
    other edge. All edge weights are equal to 1.

    Parameters
    ----------
    N : int
        Number of nodes (default is 100).
    p : float
        Probability to connect a node with another one.
    directed : bool
        Allow directed edges if True (default is False).
    self_loops : bool
        Allow self loops if True (default is False).
    connected : bool
        Force the graph to be connected (default is False).
    n_try : int
        Maximum number of trials to get a connected graph (default is 10).
    seed : int
        Seed for the random number generator (for reproducible graphs).

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> G = graphs.ErdosRenyi(N=64, seed=42)
    >>> G.set_coordinates(kind='spring', seed=42)
    >>> fig, axes = plt.subplots(1, 2)
    >>> _ = axes[0].spy(G.W, markersize=2)
    >>> _ = G.plot(ax=axes[1])

    """

    def __init__(self, N=100, p=0.1, directed=False, self_loops=False,
                 connected=False, n_try=10, seed=None, **kwargs):

        super(ErdosRenyi, self).__init__(N=N, k=1, p=p,
                                         directed=directed,
                                         self_loops=self_loops,
                                         connected=connected,
                                         n_try=n_try,
                                         seed=seed,
                                         **kwargs)



# -*- coding: utf-8 -*-

import numpy as np
from scipy import sparse

from . import Graph  # prevent circular import in Python < 3.5


class Path(Graph):
    r"""Path graph.

    A signal on the path graph is akin to a 1-dimensional signal in classical
    signal processing.

    On the path graph, the graph Fourier transform (GFT) is the classical
    discrete cosine transform (DCT_).
    As the type-II DCT, the GFT assumes even boundary conditions on both sides.

    .. _DCT: https://en.wikipedia.org/wiki/Discrete_cosine_transform

    Parameters
    ----------
    N : int
        Number of vertices.

    See Also
    --------
    Ring : 1D line with periodic boundary conditions
    Grid2d : Kronecker product of two path graphs
    Comet : Generalization with a star at one end

    References
    ----------
    :cite:`strang1999dct` shows that each DCT basis contains the eigenvectors
    of a symmetric "second difference" matrix.
    They get the eight types of DCTs by varying the boundary conditions.

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> fig, axes = plt.subplots(2, 2, figsize=(10, 8))
    >>> for i, directed in enumerate([False, True]):
    ...     G = graphs.Path(N=10, directed=directed)
    ...     _ = axes[i, 0].spy(G.W)
    ...     _ = G.plot(ax=axes[i, 1])

    The GFT of the path graph is the classical DCT.

    >>> from matplotlib import pyplot as plt
    >>> n_eigenvectors = 4
    >>> graph = graphs.Path(30)
    >>> fig, axes = plt.subplots(1, 2)
    >>> graph.set_coordinates('line1D')
    >>> graph.compute_fourier_basis()
    >>> _ = graph.plot(graph.U[:, :n_eigenvectors], ax=axes[0])
    >>> _ = axes[0].legend(range(n_eigenvectors))
    >>> _ = axes[1].plot(graph.e, '.')

    """

    def __init__(self, N=16, directed=False, **kwargs):

        self.directed = directed
        if directed:
            sources = np.arange(0, N-1)
            targets = np.arange(1, N)
            n_edges = N - 1
        else:
            sources = np.concatenate((np.arange(0, N-1), np.arange(1, N)))
            targets = np.concatenate((np.arange(1, N), np.arange(0, N-1)))
            n_edges = 2 * (N - 1)
        weights = np.ones(n_edges)
        W = sparse.csr_matrix((weights, (sources, targets)), shape=(N, N))
        plotting = {"limits": np.array([-1, N, -1, 1])}

        super(Path, self).__init__(W, plotting=plotting, **kwargs)

        self.set_coordinates('line2D')

    def _get_extra_repr(self):
        return dict(directed=self.directed)



# -*- coding: utf-8 -*-

from __future__ import division

import numpy as np

from pygsp2 import utils
from . import Filter  # prevent circular import in Python < 3.5


class MexicanHat(Filter):
    r"""Design a filter bank of Mexican hat wavelets.

    The Mexican hat wavelet is the second oder derivative of a Gaussian. Since
    we express the filter in the Fourier domain, we find:

    .. math:: \hat{g}_b(x) = x * \exp(-x)

    for the band-pass filter. Note that in our convention the eigenvalues of
    the Laplacian are equivalent to the square of graph frequencies,
    i.e. :math:`x = \lambda^2`.

    The low-pass filter is given by

    .. math: \hat{g}_l(x) = \exp(-x^4).

    Parameters
    ----------
    G : graph
    Nf : int
        Number of filters to cover the interval [0, lmax].
    lpfactor : float
        Low-pass factor. lmin=lmax/lpfactor will be used to determine scales.
        The scaling function will be created to fill the low-pass gap.
    scales : array_like
        Scales to be used.
        By default, initialized with :func:`pygsp2.utils.compute_log_scales`.
    normalize : bool
        Whether to normalize the wavelet by the factor ``sqrt(scales)``.

    Examples
    --------

    Filter bank's representation in Fourier and time (ring graph) domains.

    >>> import matplotlib.pyplot as plt
    >>> G = graphs.Ring(N=20)
    >>> G.estimate_lmax()
    >>> G.set_coordinates('line1D')
    >>> g = filters.MexicanHat(G)
    >>> s = g.localize(G.N // 2)
    >>> fig, axes = plt.subplots(1, 2)
    >>> _ = g.plot(ax=axes[0])
    >>> _ = G.plot(s, ax=axes[1])

    """

    def __init__(self, G, Nf=6, lpfactor=20, scales=None, normalize=False):

        self.lpfactor = lpfactor
        self.normalize = normalize

        lmin = G.lmax / lpfactor

        if scales is None:
            scales = utils.compute_log_scales(lmin, G.lmax, Nf-1)
        self.scales = scales

        if len(scales) != Nf - 1:
            raise ValueError('len(scales) should be Nf-1.')

        def band_pass(x):
            return x * np.exp(-x)

        def low_pass(x):
            return np.exp(-x**4)

        kernels = [lambda x: 1.2 * np.exp(-1) * low_pass(x / 0.4 / lmin)]

        for i in range(Nf - 1):

            def kernel(x, i=i):
                norm = np.sqrt(scales[i]) if normalize else 1
                return norm * band_pass(scales[i] * x)

            kernels.append(kernel)

        super(MexicanHat, self).__init__(G, kernels)

    def _get_extra_repr(self):
        return dict(lpfactor='{:.2f}'.format(self.lpfactor),
                    normalize=self.normalize)



r"""
The :mod:`pygsp2.graph_learning` module provides functions
to estimate a graph based on a distance matrix Z.

.. autosummary::

    graph_log_degree

"""
import numpy as np
from scipy.spatial.distance import squareform
from pygsp2.utils import sum_squareform


def graph_log_degree(Z, a=1.0, b=1.0, w_0='zeros', w_max=np.inf, tol=1e-5,
                     maxiter=1000, gamma=0.04, verbose=True):
    r"""Learn graph from pairwise distances using negative log prior on nodes
    degrees.

    The minimization problem solved is:

    ..math::
    minimize_W sum(sum(W * Z)) - a * sum(log(sum(W))) + b * ||W||_F^2/2 + c * ||W-W_0||_F^2/2

    Parameters
    ----------
    Z : array
        Matrix with squared pairwise distances of nodes.
    a : float
        Log prior constant (larger a -> bigger weiights in W).
    b : float
        W||_F^2 prior constant (larger b -> W denser).
    w_0 : string or matrix
        String 'zeros' will set the w priors to be a 0 vector.  Otherwise a
        matrix with the priors can be passed.
    w_max : int or float
        Maximum value for the estimated W matrix.
    tol : float
        Tolerance to end the iteration.
    maxiter : int
        Maximum number of iterations.
    gamma : float
        Step size. Number between (0,1).
    verbose : bool
        If `True` print the number of iterations.

    Returns
    -------
    W: array
        Weighted adjacency matrix.

    References
    ----------
    Vassilis Kalofolias Proceedings of the 19th International Conference on
    Artificial Intelligence and Statistics, PMLR 51:920-929, 2016.

    Examples
    -------

    Create a graph

    >>> G = graphs.Graph([[0, 0, 0, 0],
                  [0, 0, 1, 1],
                  [0, 1, 0, 1],
                  [0, 1, 1, 0]])
    >>> signal = np.array([0, 1, 1, 1])

    Compute distance

    >>> kdt = spatial.KDTree(signal[:, None])
    >>> D, NN = kdt.query(signal[:, None], k=len(signal))

    Allocate distance array
    >>> Z = np.zeros((G.N, G.N))
    >>> for i, n in enumerate(NN):
    >>>    Z[i, n] = D[i]

    Learn graph
    >>> A = 0.2
    >>> B = 0.01

    >>> W = graph_learning.graph_log_degree(Z, A, B, w_max=1)

    Threshold adjacency matrix
    >>> W[W < 1e-1] = 0

    >>> plt.subplot(121)
    >>> plt.imshow(G.W.toarray())
    >>> plt.colorbar()

    >>> plt.subplot(122)
    >>> plt.imshow(W)
    >>> plt.colorbar()

    """

    # Transform to vector space
    z = squareform(Z)
    z = z / np.amax(z)

    N = len(Z)
    S, St = sum_squareform(N)

    if w_0 == 'zeros':
        w_0 = squareform(np.zeros_like(Z))
    elif isinstance(w_0, np.ndarray) and len(w_0.shape) == 2:
        w_0 = squareform(w_0)
    else:
        raise TypeError(
            "w_0 is not 'zeros' or a 2d numpy array. Check parameter type.")

    w = w_0.copy()
    v_n = S @ w

    for i in np.arange(maxiter):

        Y = w - gamma * (2 * b * w + St @ v_n)
        y = v_n + gamma * (S @ w)

        P = np.minimum(w_max, np.maximum(0, Y - 2 * gamma * z))
        p = y - gamma * a * \
            ((y/(gamma * a)) + np.sqrt(((y/(gamma*a))**2) + (4/(gamma*a))))/2

        Q = P - gamma * (2 * b * P + St @ p)
        q = p + gamma * (S @ P)

        w_new = w - Y + Q
        v_new = v_n - y + q

        w = w_new.copy()
        v_n = v_new.copy()

        if (np.linalg.norm(- Y + Q) / np.linalg.norm(w) < tol) and \
                (np.linalg.norm(- y + q) / np.linalg.norm(v_n) < tol):
            break

    if verbose:
        print(f'Found solution after {i+1} iterations')

    return squareform(w)



# -*- coding: utf-8 -*-

import numpy as np
from scipy import sparse

from . import Graph  # prevent circular import in Python < 3.5


class Ring(Graph):
    r"""K-regular ring graph.

    A signal on the ring graph is akin to a 1-dimensional periodic signal in
    classical signal processing.

    On the ring graph, the graph Fourier transform (GFT) is the classical
    discrete Fourier transform (DFT_).
    Actually, the Laplacian of the ring graph is a `circulant matrix`_, and any
    circulant matrix is diagonalized by the DFT.

    .. _DFT: https://en.wikipedia.org/wiki/Discrete_Fourier_transform
    .. _circulant matrix: https://en.wikipedia.org/wiki/Circulant_matrix

    Parameters
    ----------
    N : int
        Number of vertices.
    k : int
        Number of neighbors in each direction.

    See Also
    --------
    Path : 1D line with even boundary conditions
    Torus : Kronecker product of two ring graphs

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> G = graphs.Ring(N=10)
    >>> fig, axes = plt.subplots(1, 2)
    >>> _ = axes[0].spy(G.W)
    >>> _ = G.plot(ax=axes[1])

    The GFT of the ring graph is the classical DFT.

    >>> from matplotlib import pyplot as plt
    >>> n_eigenvectors = 4
    >>> graph = graphs.Ring(30)
    >>> fig, axes = plt.subplots(1, 2)
    >>> graph.set_coordinates('line1D')
    >>> graph.compute_fourier_basis()
    >>> _ = graph.plot(graph.U[:, :n_eigenvectors], ax=axes[0])
    >>> _ = axes[0].legend(range(n_eigenvectors))
    >>> _ = axes[1].plot(graph.e, '.')

    """

    def __init__(self, N=64, k=1, **kwargs):

        self.k = k

        if N < 3:
            # Asymmetric graph needed for 2 as 2 distances connect them.
            raise ValueError('There should be at least 3 vertices.')

        if 2*k > N:
            raise ValueError('Too many neighbors requested.')

        if 2*k == N:
            num_edges = N * (k - 1) + k
        else:
            num_edges = N * k

        i_inds = np.zeros((2 * num_edges))
        j_inds = np.zeros((2 * num_edges))

        tmpN = np.arange(N, dtype=int)
        for i in range(min(k, (N - 1) // 2)):
            i_inds[2*i * N + tmpN] = tmpN
            j_inds[2*i * N + tmpN] = np.remainder(tmpN + i + 1, N)
            i_inds[(2*i + 1)*N + tmpN] = np.remainder(tmpN + i + 1, N)
            j_inds[(2*i + 1)*N + tmpN] = tmpN

        if 2*k == N:
            i_inds[2*N*(k - 1) + tmpN] = tmpN
            i_inds[2*N*(k - 1) + tmpN] = np.remainder(tmpN + k + 1, N)

        W = sparse.csc_matrix((np.ones((2*num_edges)), (i_inds, j_inds)),
                              shape=(N, N))

        plotting = {'limits': np.array([-1, 1, -1, 1])}

        super(Ring, self).__init__(W, plotting=plotting, **kwargs)

        self.set_coordinates('ring2D')

    def _get_extra_repr(self):
        return dict(k=self.k)



# -*- coding: utf-8 -*-

r"""
The :mod:`pygsp2.graphs` module implements the graph class hierarchy. A graph
object is either constructed from an adjacency matrix, or by instantiating one
of the built-in graph models.

Interface
=========

The :class:`Graph` base class allows to construct a graph object from any
adjacency matrix and provides a common interface to that object. Derived
classes then allows to instantiate various standard graph models.

Attributes
----------

**Matrix operators**

.. autosummary::

    Graph.W
    Graph.L
    Graph.U
    Graph.D

**Vectors**

.. autosummary::

    Graph.d
    Graph.dw
    Graph.e

**Scalars**

.. autosummary::

    Graph.lmax
    Graph.coherence

Attributes computation
----------------------

.. autosummary::

    Graph.compute_laplacian
    Graph.estimate_lmax
    Graph.compute_fourier_basis
    Graph.compute_differential_operator

Differential operators
----------------------

.. autosummary::

    Graph.grad
    Graph.div
    Graph.dirichlet_energy

Transforms
----------

.. autosummary::

    Graph.gft
    Graph.igft

Vertex-frequency transforms are implemented as filter banks and are found in
:mod:`pygsp2.filters` (such as :class:`~pygsp2.filters.Gabor` and
:class:`~pygsp2.filters.Modulation`).

Checks
------

.. autosummary::

    Graph.is_weighted
    Graph.is_connected
    Graph.is_directed
    Graph.has_loops

Plotting
--------

.. autosummary::

    Graph.plot
    Graph.plot_spectrogram

Import and export (I/O)
-----------------------

We provide import and export facility to two well-known Python packages for
network analysis: NetworkX_ and graph-tool_.
Those packages and the PyGSP2 are fundamentally different in their goals (graph
analysis versus graph signal analysis) and graph representations (if in the
PyGSP2 everything is an ndarray, in NetworkX everything is a dictionary).
Those tools are complementary and good interoperability is necessary to exploit
the strengths of each tool.
We ourselves leverage NetworkX and graph-tool to save and load graphs.

Note: to tie a signal with the graph, such that they are exported together,
attach it first with :meth:`Graph.set_signal`.

.. _NetworkX: https://networkx.org
.. _graph-tool: https://graph-tool.skewed.de

.. autosummary::

    Graph.load
    Graph.save
    Graph.from_networkx
    Graph.to_networkx
    Graph.from_graphtool
    Graph.to_graphtool

Others
------

.. autosummary::

    Graph.get_edge_list
    Graph.set_signal
    Graph.set_coordinates
    Graph.subgraph
    Graph.extract_components

Graph models
============

In addition to the below graphs, useful resources are the random graph
generators from NetworkX (see `NetworkX's documentation`_) and graph-tool (see
:mod:`graph_tool.generation`), as well as graph-tool's assortment of standard
networks (see :mod:`graph_tool.collection`).
Any graph created by NetworkX or graph-tool can be imported in the PyGSP2 with
:meth:`Graph.from_networkx` and :meth:`Graph.from_graphtool`.

.. _NetworkX's documentation: https://networkx.org/documentation/stable/reference/generators.html

Graphs built from other graphs
------------------------------

.. autosummary::

    LineGraph

Generated graphs
----------------

.. autosummary::

    Airfoil
    BarabasiAlbert
    Comet
    Community
    DavidSensorNet
    ErdosRenyi
    FullConnected
    Grid2d
    Logo
    LowStretchTree
    Minnesota
    Path
    RandomRegular
    RandomRing
    Ring
    Star
    StochasticBlockModel
    SwissRoll
    Torus

Nearest-neighbors graphs constructed from point clouds
------------------------------------------------------

.. autosummary::

    NNGraph
    Bunny
    Cube
    ImgPatches
    Grid2dImgPatches
    Sensor
    Sphere
    TwoMoons

"""

from .graph import Graph  # noqa: F401
from .airfoil import Airfoil  # noqa: F401
from .barabasialbert import BarabasiAlbert  # noqa: F401
from .comet import Comet  # noqa: F401
from .community import Community  # noqa: F401
from .davidsensornet import DavidSensorNet  # noqa: F401
from .erdosrenyi import ErdosRenyi  # noqa: F401
from .fullconnected import FullConnected  # noqa: F401
from .grid2d import Grid2d  # noqa: F401
from .linegraph import LineGraph  # noqa: F401
from .logo import Logo  # noqa: F401
from .lowstretchtree import LowStretchTree  # noqa: F401
from .minnesota import Minnesota  # noqa: F401
from .path import Path  # noqa: F401
from .randomregular import RandomRegular  # noqa: F401
from .randomring import RandomRing  # noqa: F401
from .ring import Ring  # noqa: F401
from .star import Star  # noqa: F401
from .stochasticblockmodel import StochasticBlockModel  # noqa: F401
from .swissroll import SwissRoll  # noqa: F401
from .torus import Torus  # noqa: F401

from .nngraphs.nngraph import NNGraph  # noqa: F401
from .nngraphs.bunny import Bunny  # noqa: F401
from .nngraphs.cube import Cube  # noqa: F401
from .nngraphs.imgpatches import ImgPatches  # noqa: F401
from .nngraphs.grid2dimgpatches import Grid2dImgPatches  # noqa: F401
from .nngraphs.sensor import Sensor  # noqa: F401
from .nngraphs.sphere import Sphere  # noqa: F401
from .nngraphs.twomoons import TwoMoons  # noqa: F401



# -*- coding: utf-8 -*-

import numpy as np

from pygsp2 import utils
from . import Graph  # prevent circular import in Python < 3.5


class DavidSensorNet(Graph):
    r"""Sensor network.

    Parameters
    ----------
    N : int
        Number of vertices (default = 64). Values of 64 and 500 yield
        pre-computed and saved graphs. Other values yield randomly generated
        graphs.
    seed : int
        Seed for the random number generator (for reproducible graphs).

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> G = graphs.DavidSensorNet()
    >>> fig, axes = plt.subplots(1, 2)
    >>> _ = axes[0].spy(G.W, markersize=2)
    >>> _ = G.plot(ax=axes[1])

    """

    def __init__(self, N=64, seed=None, **kwargs):

        self.seed = seed

        if N == 64:
            data = utils.loadmat('pointclouds/david64')
            assert data['N'][0, 0] == N
            W = data['W']
            coords = data['coords']

        elif N == 500:
            data = utils.loadmat('pointclouds/david500')
            assert data['N'][0, 0] == N
            W = data['W']
            coords = data['coords']

        else:
            coords = np.random.default_rng(seed).uniform(size=(N, 2))

            target_dist_cutoff = -0.125 * N / 436.075 + 0.2183
            T = 0.6
            s = np.sqrt(-target_dist_cutoff**2/(2*np.log(T)))
            d = utils.distanz(coords.T)
            W = np.exp(-np.power(d, 2)/(2.*s**2))
            W[W < T] = 0
            W[np.diag_indices(N)] = 0

        plotting = {"limits": [0, 1, 0, 1]}

        super(DavidSensorNet, self).__init__(W, coords=coords,
                                             plotting=plotting, **kwargs)

    def _get_extra_repr(self):
        return dict(seed=self.seed)



r"""
Metro simulation 2
==================

Example of a simulation of the evolution of a graph signal over the Santiago Metro
network.

Simulation of the distribution of a signal over the metro network. The starting
conditions is a graph signal with only one positive integer value bigger.The
plots or animation show how this signal distributes over the network using:

.. math:: y = AD^-1x

This example adds some constraints to the previous metro simulation
example. Some of them are:

1. Eliminates the backward connections on the simulation and substracts 1 to
    every node degree to avoid dilution of the signal by going backwards.

2. If the 2 wagons arrive at the same metro station, there is no other way but
    out. All the people in those wagons will exit the sations.

3. In terminal stations, the people will not travel backwards this will also
    force an exit to everyone that reaches the terminal station.

4. If there are less than the number of people that is set to get out in every
    station then then wagon is emptyed.

The output is a folder with numerated figures. You can use websites such as
https://gifmaker.me/ to make an animation with the resulting figures.

The initial condition is a graph signal with only one positive integer value
bigger.The plots or animation show how this signal distributes over the network
using:

.. math:: y = AD^-1x

To run this example, you need to download three files and place them in the
same directory as this script.

1. Download the file `Tablas de subidas y bajadas nov23.zip` from this link:

https://www.dtpm.cl/descargas/modelos_y_matrices/Tablas%20de%20subidas%20y%20bajadas%20nov23.zip

Then, uncompress the zip file and copy `2023.11 Matriz_baj_SS_MH.xlsb` to the
same location as this script.

2. Download the file `santiago_metro_stations_coords.geojson` from this link:

https://zenodo.org/records/11637462/files/santiago_metro_stations_coords.geojson

3. Download the file `santiago_metro_stations_connections.txt` from this link:

https://zenodo.org/records/11637462/files/santiago_metro_stations_connections.txt
"""
# %%
import os
import numpy as np
import networkx as nx
import matplotlib.pyplot as plt
from pygsp2.utils_examples import make_metro_graph, fetch_data

current_dir = os.getcwd()
os.chdir(current_dir)

try:
    os.mkdir('metro_simulation2/')
except FileExistsError:
    print(
        'Warning: It seems like this folder already exists. Overwritting...')

# %% Download data
assets_dir = os.path.join(current_dir, 'data')
fetch_data(assets_dir, "metro")
# %% Load graph and compute adjacency and node degree
G, pos = make_metro_graph(
    edgesfile=os.path.join(assets_dir, 'santiago_metro_stations_connections.txt'),
    coordsfile=os.path.join(assets_dir, 'santiago_metro_stations_coords.geojson')
)
stations = list(G)

W0 = nx.adjacency_matrix(G).toarray()
D = np.diag(W0@np.ones(len(W0)))
# The signal will advance through the graph. In our case
# the backwards direction needs to be subtracted. Otherwise
# the signal will be "diluted" in each step by half since
# there is a backwards connection available.
D[D > 1] = D[D > 1] - 1
D_inv = np.linalg.inv(D)

# Store terminal stations. Anything that arrives here
# will get out of the network.
terminal_stations = np.where(np.sum(W0, axis=1) == 1)


# %% Initialize parameters
NSTEPS = 27  # Arbitrary units, steps
INIT_VALUE = 5000  # Initial conditions
INIT_STATION = 103  # HOSPITAL DEL PINO
# Set constant value that will exit
# the network in each step
OUT_CONSTANT = 10

signal = np.zeros(len(W0))
signal[INIT_STATION] = INIT_VALUE

# %% Initialize graph
normalized_signal = signal / INIT_VALUE
nodelist = [list(G)[INIT_STATION]]

fig, ax = plt.subplots(figsize=(10, 7))
cmap = plt.get_cmap('viridis')
colors = cmap(normalized_signal[signal > 0])

nx.draw_networkx_edges(G, pos, node_size=20, ax=ax)
im = nx.draw_networkx_nodes(G, pos, node_color='gray', node_size=20, ax=ax)
im = nx.draw_networkx_nodes(
    G, pos, node_color=colors, nodelist=nodelist, node_size=20, ax=ax)
cbar = plt.colorbar(im, ax=ax, label='Number of people',
                    ticks=[0, 0.5, 1])
cbar.set_ticklabels([0, (OUT_CONSTANT*10)/2, (OUT_CONSTANT*10)])
plt.savefig('metro_simulation2/0.png')

# %% Perform simulation

visited_stations = [INIT_STATION]
mask = np.ones_like(W0)
W = W0 * mask

# People that leave the
# metro network in each station
left = np.zeros_like(signal)

for it in np.arange(1, NSTEPS):

    # Compute signal in new step
    signal = (W@D_inv@signal).astype(int)

    # Eliminate connections of visited stations
    mask[visited_stations, :] = 0
    mask[:, visited_stations] = 0

    W = W0 * mask

    # Check if people reached a terminal station
    stations2empty = np.intersect1d(np.where(signal)[0], terminal_stations)
    # Check if people reached a station that has no connections
    stations2empty2 = np.intersect1d(
        np.where(signal > 0)[0], np.where(W.sum(1) == 0)[0])

    if stations2empty.size > 0:
        print(f'It {it}: Emptying terminal stations')
        for i in stations2empty:
            print(f'\t{signal[i]} unboarded {stations[i]}')
            left[i] += signal[i]
            signal[i] -= signal[i]
    elif stations2empty2.size > 0:
        print(f'It {it}: Emptying encountered stations')
        for i in stations2empty2:
            print(f'\t{signal[i]} unboarded {stations[i]}')
            left[i] += signal[i]
            signal[i] -= signal[i]

    # Check if there is enough people to get out, otherwise
    # empty the station
    if np.all(signal[np.where(signal)[0]] > OUT_CONSTANT):
        for i in np.where(signal > 0)[0]:
            left[i] += OUT_CONSTANT
            signal[i] -= OUT_CONSTANT
    else:
        print(f'It {it}: Emptying stations')
        for i in np.where(np.logical_and(signal > 0, signal < OUT_CONSTANT))[0]:
            print(f'\t{signal[i]} unboarded {stations[i]}')
            left[i] += signal[i]
            signal[i] -= signal[i]
        for i in np.where(signal > 0)[0]:
            left[i] += OUT_CONSTANT
            signal[i] -= OUT_CONSTANT

    current_station = np.where(signal > 0)[0]
    nodelist = [list(G)[i] for i in current_station]
    nodelist2 = [list(G)[i] for i in np.where(left)[0]]

    normalized_signal = signal / INIT_VALUE
    colors = cmap(normalized_signal[signal > 0])

    normalized_left = left / (OUT_CONSTANT*10)
    colors2 = cmap(normalized_left[left > 0])

    # Update visited stations list
    for i in np.where(signal)[0]:
        visited_stations.append(i)

    fig, ax = plt.subplots(figsize=(10, 7))

    nx.draw_networkx_edges(G, pos, node_size=20, ax=ax)
    im = nx.draw_networkx_nodes(G, pos, node_color='gray', node_size=20, ax=ax)

    im = nx.draw_networkx_nodes(
        G, pos, node_color=colors2, nodelist=nodelist2, node_size=20, ax=ax)

    im = nx.draw_networkx_nodes(
        G, pos, node_color=colors, nodelist=nodelist, node_size=20, ax=ax,
        node_shape='o', edgecolors='red')

    cbar = plt.colorbar(im, ax=ax, label='Number of people',
                        ticks=[0, 0.5, 1])
    cbar.set_ticklabels([0, (OUT_CONSTANT*10)/2, (OUT_CONSTANT*10)])

    plt.savefig(f'metro_simulation2/{it}.png')
    plt.close()
    # fig.clf()

    if (np.sum(signal) == 0) or (np.sum(left) == INIT_VALUE):
        print(f'It {it}: Finished...')
        break
